import functools

from django.contrib.admin.utils import (
    get_fields_from_path, reverse_field_path
)
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from .exceptions import DynamicSearchException, DynamicSearchModelException
from .literals import QUERY_PARAMETER_ANY_FIELD
from .value_transformations import ValueTransformation


class SearchField:
    _registry = []
    _registry_by_class = {}

    class_label = None
    # Collection is a field that return multiple related fields as a
    # queryset.
    collection = None
    # Concrete is true if the search fields maps 1 to 1 with a real model
    # field.
    concrete = None
    # Priority ensures concrete and collection fields are processed before
    # virtual fields.
    priority = None

    @staticmethod
    class ValueTransformationNull(ValueTransformation):
        def _execute(self):
            return self.value

    @classmethod
    def all(cls):
        return cls._registry

    @classmethod
    def get_by_class(cls, klass):
        return cls._registry_by_class[klass]

    @classmethod
    def get_classes(cls):
        return cls._registry

    @classmethod
    def init(cls, *args, **kwargs):
        for klass in cls.all():
            if klass.check(*args, **kwargs):
                return klass(*args, **kwargs)

        raise DynamicSearchException(
            'No available `SearchField` subclass for this model field.'
        )

    @classmethod
    def register(cls, klass):
        cls._registry.append(klass)

    def __init__(self, field, search_model, help_text=None, label=None):
        self._label = label
        self._field_name = field
        self.help_text = help_text
        self.search_model = search_model

        self.__class__._registry_by_class.setdefault(
            self.__class__, []
        )
        self.__class__._registry_by_class[self.__class__].append(self)

    def __repr__(self):
        return '<{}: {}.{}>'.format(
            self.__class__.__name__, self.model._meta.label,
            self.field_name
        )

    def __str__(self):
        return '{}.{}'.format(
            self.model._meta.label, self.field_name
        )

    def _get_class_registry_index(self):
        return self.__class__._registry.index(self.__class__)

    @cached_property
    def field_class(self):
        return self.model_field.__class__

    @cached_property
    def field_class_label(self):
        return self.field_class.__name__

    @cached_property
    def field_name(self):
        return self._field_name

    field_name.help_text = _(message='Uniquely identifies the search field.')
    field_name.short_description = _(message='Field name')

    @cached_property
    def field_name_model_list(self):
        result = []

        model_list = get_fields_from_path(
            model=self.model, path=self.field_name
        )

        for model in model_list:
            remote_field = model.remote_field

            if remote_field:
                model = remote_field.model
            else:
                model = self.search_model.model

            result.append(model)

        return result

    def do_value_index_transform(self, search_backend, value):
        return self.do_value_transform(
            key='index', search_backend=search_backend, value=value
        )

    def do_value_search_transform(self, search_backend, value):
        return self.do_value_transform(
            key='search', search_backend=search_backend, value=value
        )

    def do_value_transform(self, key, search_backend, value):
        transformations = self.get_backend_field_transformations(
            search_backend=search_backend
        )

        transformation_class_list = transformations.get(
            key, (SearchField.ValueTransformationNull,)
        )

        for transformation_class in transformation_class_list:
            value = transformation_class(value=value).execute()

        return value

    @functools.cache
    def get_backend_field_query_type_list(self, search_backend):
        return self.get_search_field_type_backend_dictionary(
            search_backend=search_backend
        ).get(
            'query_type_list', []
        )

    @functools.cache
    def get_backend_field_transformations(self, search_backend):
        return self.get_search_field_type_backend_dictionary(
            search_backend=search_backend
        ).get(
            'transformations', {}
        )

    def get_help_text(self):
        return self.help_text or getattr(self.model_field, 'help_text', '')

    def get_label(self):
        return self.label or self.field_name

    def get_model_field(self):
        return get_fields_from_path(
            model=self.model, path=self.field_name
        )[-1]

    def get_search_field_class_label(self):
        cls = self.__class__
        return cls.class_label or cls.__name__

    get_search_field_class_label.help_text = _(
        message='The type of search field. This determines how the '
        'information will be handled for search indexing.'
    )
    get_search_field_class_label.short_description = _(
        message='Search field class'
    )

    @functools.cache
    def get_search_field_type_backend_dictionary(self, search_backend):
        try:
            search_backend_field_type_attributes = search_backend.get_field_type_mapping()[
                self.field_class
            ]
        except KeyError:
            return {}
        else:
            return search_backend_field_type_attributes

    @property
    def label(self):
        return self._label or self.model_field.verbose_name

    @cached_property
    def model(self):
        return self.search_model.model

    @cached_property
    def model_field(self):
        return self.get_model_field()

    @cached_property
    def related_model(self):
        return self.model_field.model

    @cached_property
    def reverse_path(self):
        parent, path = reverse_field_path(
            model=self.model, path=self.field_name
        )

        return path


class SearchFieldConcrete(SearchField):
    class_label = _('Concrete')
    concrete = True

    def _do_field_name_verify(self):
        try:
            self.model_field
        except FieldDoesNotExist:
            raise DynamicSearchModelException(
                'Unknown field or field path `{}`.'.format(self.field_name)
            )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._do_field_name_verify()


class SearchFieldDirect(SearchFieldConcrete):
    """
    Search for terms in fields that directly belong to the parent
    SearchModel.
    """

    class_label = _('Direct')
    collection = False
    priority = 0

    @classmethod
    def check(cls, *args, **kwargs):
        return LOOKUP_SEP not in kwargs['field'] and not kwargs['field'] == QUERY_PARAMETER_ANY_FIELD

    def get_instance_value(
        self, instance, search_backend, exclude_kwargs=None,
        exclude_model=None, instance_field_data=None
    ):
        value = getattr(instance, self.field_name)

        value = self.do_value_index_transform(
            search_backend=search_backend, value=value
        )

        return search_backend.do_native_type_conversion(value=value)


class SearchFieldRelated(SearchFieldConcrete):
    class_label = _('Related')
    collection = True
    priority = 1

    @classmethod
    def check(cls, *args, **kwargs):
        return LOOKUP_SEP in kwargs['field'] and not kwargs['field'] == QUERY_PARAMETER_ANY_FIELD

    def get_instance_value(
        self, instance, search_backend, exclude_kwargs=None,
        exclude_model=None, instance_field_data=None
    ):
        last_field = self.field_name.split(LOOKUP_SEP)[-1]

        sub_queryset = self.related_model._meta.default_manager.filter(
            **{
                '{}'.format(self.reverse_path): instance.pk
            }
        ).values_list(last_field, flat=True)

        sub_queryset = sub_queryset.filter(
            **{
                '{field_name}{lookup_separator}isnull'.format(
                    field_name=last_field, lookup_separator=LOOKUP_SEP
                ): False
            }
        )

        if exclude_model and self.related_model == exclude_model:
            sub_queryset = sub_queryset.exclude(**exclude_kwargs)

        result = []

        sub_queryset = sub_queryset.distinct()

        for item in sub_queryset:
            item_value = self.do_value_index_transform(
                search_backend=search_backend, value=item
            )
            if item_value:
                result.append(item_value)

        return search_backend.do_native_type_conversion(value=result)


class SearchFieldVirtual(SearchField):
    """
    Base class for all virtual search field that are not populated by
    directly reading model instance attributes but a calculation.

    Also cover search fields that might not translate to a single model
    field.

    Virtual search fields need to be registered after the other search
    field in order to have access to their instance data when doing
    model instance indexing.
    """

    class_label = _('Virtual')
    collection = False
    concrete = False
    priority = 2

    @cached_property
    def field_class(self):
        return models.TextField

    def get_model_field(self):
        field_list = get_fields_from_path(model=self.model, path='id')
        return field_list[-1]


class SearchFieldVirtualAllFields(SearchFieldVirtual):
    label = _(message='Any')

    @classmethod
    def check(cls, *args, **kwargs):
        return kwargs['field'] == QUERY_PARAMETER_ANY_FIELD

    @cached_property
    def field_composition(self):
        return [
            search_field for search_field in self.search_model.search_fields if search_field.concrete
        ]

    def get_instance_value(
        self, instance, search_backend, exclude_model=None,
        exclude_kwargs=None, instance_field_data=None
    ):
        return None


SearchField.register(klass=SearchFieldDirect)
SearchField.register(klass=SearchFieldRelated)
SearchField.register(klass=SearchFieldVirtualAllFields)



from django.utils.translation import gettext_lazy as _

from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import MultiKwargHyperlinkedIdentityField

from ..models.document_file_models import DocumentFile
from ..models.document_file_page_models import DocumentFilePage


class DocumentFilePageSerializer(serializers.HyperlinkedModelSerializer):
    document_file_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Document file URL'), view_kwargs=(
            {
                'lookup_field': 'document_file.document.pk',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'document_file_id',
                'lookup_url_kwarg': 'document_file_id'
            }
        ), view_name='rest_api:documentfile-detail'
    )
    image_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Image URL'), view_kwargs=(
            {
                'lookup_field': 'document_file.document.pk',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'document_file_id',
                'lookup_url_kwarg': 'document_file_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'document_file_page_id'
            }
        ), view_name='rest_api:documentfilepage-image'
    )
    url = MultiKwargHyperlinkedIdentityField(
        label=_(message='URL'), view_kwargs=(
            {
                'lookup_field': 'document_file.document.pk',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'document_file_id',
                'lookup_url_kwarg': 'document_file_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'document_file_page_id'
            }
        ), view_name='rest_api:documentfilepage-detail'
    )

    class Meta:
        fields = (
            'document_file_id', 'document_file_url', 'id', 'image_url',
            'page_number', 'url'
        )
        model = DocumentFilePage
        read_only_fields = (
            'document_file_id', 'document_file_url', 'id', 'image_url', 'url'
        )


class DocumentFileSerializer(serializers.HyperlinkedModelSerializer):
    action_name = serializers.CharField(
        label=_(message='Action name'), write_only=True
    )
    document_url = serializers.HyperlinkedIdentityField(
        label=_(message='Document URL'), lookup_field='document_id',
        lookup_url_kwarg='document_id', view_name='rest_api:document-detail'
    )
    file_new = serializers.FileField(
        help_text=_(message='Binary content for the new file.'),
        label=_(message='File new'), use_url=False, write_only=True
    )
    page_list_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Page list URL'), view_kwargs=(
            {
                'lookup_field': 'document_id',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'document_file_id'
            }
        ), view_name='rest_api:documentfilepage-list'
    )
    pages_first = DocumentFilePageSerializer(
        label=_(message='Pages first'), many=False, read_only=True
    )
    url = MultiKwargHyperlinkedIdentityField(
        label=_(message='URL'), view_kwargs=(
            {
                'lookup_field': 'document_id',
                'lookup_url_kwarg': 'document_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'document_file_id'
            }
        ),
        view_name='rest_api:documentfile-detail'
    )

    class Meta:
        create_only_fields = ('action_name', 'file_new',)
        extra_kwargs = {
            'file': {'use_url': False}
        }
        fields = (
            'action_name', 'checksum', 'comment', 'document_id',
            'document_url', 'encoding', 'file', 'file_new', 'filename', 'id',
            'mimetype', 'page_list_url', 'pages_first', 'size', 'timestamp',
            'url'
        )
        model = DocumentFile
        read_only_fields = (
            'checksum', 'document_id', 'document_url', 'encoding', 'file',
            'id', 'mimetype', 'page_list_url', 'pages_first', 'size',
            'timestamp', 'url'
        )



from mayan.apps.navigation.column_widgets import SourceColumnWidget


class WorkflowLogExtraDataWidget(SourceColumnWidget):
    template_name = 'document_states/extra_data.html'


class WorkflowTemplateTransitionTriggerColumnWidget(SourceColumnWidget):
    template_name = 'document_states/column_widgets/transition_triggers.html'



from mayan.apps.icons.icons import Icon

# App

icon_file_caching = Icon(
    driver_name='fontawesome', symbol='warehouse'
)

# Cache
icon_cache_detail = Icon(
    driver_name='fontawesome', symbol='warehouse'
)
icon_cache_list = Icon(
    driver_name='fontawesome', symbol='warehouse'
)
icon_cache_purge_multiple = Icon(
    driver_name='fontawesome-dual', primary_symbol='warehouse',
    secondary_symbol='check'
)
icon_cache_purge_single = icon_cache_purge_multiple

# Cache partition

icon_cache_partition_detail = Icon(
    driver_name='fontawesome', symbol='warehouse'
)
icon_cache_partition_purge = Icon(
    driver_name='fontawesome-dual', primary_symbol='warehouse',
    secondary_symbol='check'
)



import logging

import sh

from django.core.files import File
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.serialization import yaml_load
from mayan.apps.source_generated_files.source_backend_actions.generated_file_actions import (
    SourceBackendActionGenerateFileDocumentFileUpload,
    SourceBackendActionGenerateFileDocumentUpload
)
from mayan.apps.source_interactive.source_backends.mixins import (
    SourceBackendMixinInteractive
)
from mayan.apps.sources.settings import setting_backend_arguments
from mayan.apps.sources.source_backends.base import SourceBackend
from mayan.apps.storage.utils import NamedTemporaryFile, touch

from .literals import DEFAULT_BINARY_SCANIMAGE_PATH

logger = logging.getLogger(name=__name__)


class SourceBackendSANEScanner(SourceBackendMixinInteractive, SourceBackend):
    action_class_list = (
        SourceBackendActionGenerateFileDocumentFileUpload,
        SourceBackendActionGenerateFileDocumentUpload
    )
    label = _(message='SANE Scanner')
    widgets = {
        'arguments': {
            'class': 'django.forms.widgets.Textarea', 'kwargs': {
                'attrs': {
                    'rows': 10
                }
            }
        }
    }

    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()
        fields.update(
            {
                'device_name': {
                    'class': 'django.forms.CharField',
                    'help_text': _(
                        message='Device name as returned by the SANE backend.'
                    ),
                    'kwargs': {'max_length': 255},
                    'label': _(message='Device name'),
                    'required': True
                },
                'arguments': {
                    'class': 'django.forms.CharField',
                    'help_text': _(
                        message='YAML formatted arguments to pass to the '
                        '`scanimage` command. The arguments will change '
                        'depending on the device. Execute `scanimage --help '
                        '--device-name=DEVICE` for the list of supported '
                        'arguments.'
                    ),
                    'label': _(message='Arguments'),
                    'required': False
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='SANE client'), {
                    'fields': ('device_name', 'arguments')
                },
            ),
        )

        return fieldsets

    def action_file_get(self, file_identifier):
        with NamedTemporaryFile() as file_object:
            # The output_file argument is only supported in version 1.0.28
            # https://gitlab.com/sane-project/backends/-/releases/1.0.28
            # Using redirection make this compatible with more versions.
            try:
                self.call_command_scanimage(_out=file_object.name)
            except sh.ErrorReturnCode:
                # The shell command is deleting the temporary file on errors.
                # Recreate it so that `NamedTemporaryFile` is able to delete
                # it when the context exits.
                touch(filename=file_object.name)
                raise
            else:
                file_object.seek(0)

                filename = 'scan {}'.format(
                    now()
                )
                yield {
                    'file': File(file=file_object, name=filename)
                }

    def call_command_scanimage(self, **kwargs):
        command_path = setting_backend_arguments.value.get(
            'mayan.apps.sources.source_backends.SourceBackendSANEScanner', {}
        ).get('scanimage_path', DEFAULT_BINARY_SCANIMAGE_PATH)

        command_scanimage = sh.Command(path=command_path)

        command_scanimage = command_scanimage.bake(
            device_name=self.kwargs['device_name'],
            format='tiff'
        )

        loaded_arguments = yaml_load(
            stream=self.kwargs.get('arguments', '{}')
        ) or {}

        loaded_arguments.update(**kwargs)

        command_scanimage(**loaded_arguments)

    def get_view_context(self, context, request):
        return {
            'subtemplates_list': [
                {
                    'context': {
                        'forms': context['forms'],
                        'is_multipart': True,
                        'title': _(message='Document properties'),
                        'submit_label': _(message='Scan')
                    },
                    'name': 'appearance/partials/form/multiple.html'
                }
            ]
        }



from django.utils.translation import gettext_lazy as _

from rest_framework.reverse import reverse

from mayan.apps.rest_api import serializers

from ..models import IndexInstance, IndexInstanceNode


class IndexInstanceSerializer(serializers.ModelSerializer):
    depth = serializers.IntegerField(
        label=_(message='Depth'), read_only=True, source='get_level_count'
    )
    node_count = serializers.IntegerField(
        label=_(message='Node count'), read_only=True,
        source='get_descendants_count'
    )
    nodes_url = serializers.SerializerMethodField(
        label=_(message='Nodes URL'), read_only=True
    )
    url = serializers.SerializerMethodField(
        label=_(message='URL'), read_only=True
    )

    class Meta:
        fields = ('depth', 'label', 'id', 'nodes_url', 'node_count', 'url')
        model = IndexInstance
        read_only_fields = fields

    def get_url(self, obj):
        return reverse(
            format=self.context['format'], kwargs={
                'index_instance_id': obj.pk
            }, request=self.context['request'],
            viewname='rest_api:indexinstance-detail'
        )

    def get_nodes_url(self, obj):
        return reverse(
            format=self.context['format'], kwargs={
                'index_instance_id': obj.pk
            }, request=self.context['request'],
            viewname='rest_api:indexinstancenode-list'
        )


class IndexInstanceNodeSerializer(serializers.ModelSerializer):
    children_url = serializers.SerializerMethodField(
        label=_(message='Children URL'), read_only=True
    )
    depth = serializers.IntegerField(
        label=_(message='Depth'), read_only=True, source='get_level_count'
    )
    documents_url = serializers.SerializerMethodField(
        label=_(message='Documents URL'), read_only=True
    )
    index_url = serializers.SerializerMethodField(
        label=_(message='Index URL'), read_only=True
    )
    node_count = serializers.IntegerField(
        label=_(message='Node count'), read_only=True,
        source='get_descendants_count'
    )
    parent_url = serializers.SerializerMethodField(
        label=_(message='Parent URL'), read_only=True
    )
    url = serializers.SerializerMethodField(
        label=_(message='URL'), read_only=True
    )

    class Meta:
        fields = (
            'depth', 'documents_url', 'children_url', 'id', 'index_url',
            'level', 'node_count', 'parent_id', 'parent_url', 'value', 'url'
        )
        model = IndexInstanceNode
        read_only_fields = fields

    def get_children_url(self, obj):
        return reverse(
            format=self.context['format'], kwargs={
                'index_instance_id': obj.index().pk,
                'index_instance_node_id': obj.pk
            }, request=self.context['request'],
            viewname='rest_api:indexinstancenode-children-list'
        )

    def get_documents_url(self, obj):
        return reverse(
            format=self.context['format'], kwargs={
                'index_instance_id': obj.index().pk,
                'index_instance_node_id': obj.pk
            }, request=self.context['request'],
            viewname='rest_api:indexinstancenode-document-list'
        )

    def get_index_url(self, obj):
        return reverse(
            format=self.context['format'], kwargs={
                'index_instance_id': obj.index().pk
            }, request=self.context['request'],
            viewname='rest_api:indexinstance-detail'
        )

    def get_parent_url(self, obj):
        if obj.parent and not obj.parent.is_root_node():
            return reverse(
                format=self.context['format'], kwargs={
                    'index_instance_id': obj.index().pk,
                    'index_instance_node_id': obj.parent.pk
                }, request=self.context['request'],
                viewname='rest_api:indexinstancenode-detail'
            )
        else:
            return ''

    def get_url(self, obj):
        return reverse(
            format=self.context['format'], kwargs={
                'index_instance_id': obj.index().pk,
                'index_instance_node_id': obj.pk
            }, request=self.context['request'],
            viewname='rest_api:indexinstancenode-detail'
        )



from django.http import Http404

from mayan.apps.rest_api import generics

from ..classes import EventType, EventTypeNamespace
from ..serializers.event_type_serializers import (
    EventTypeNamespaceSerializer, EventTypeSerializer
)


class APIEventTypeListView(generics.ListAPIView):
    """
    get: Returns a list of all the available event types.
    """
    serializer_class = EventTypeSerializer
    source_queryset = EventType.all()

    def get_serializer_context(self):
        return {
            'format': self.format_kwarg,
            'request': self.request,
            'view': self
        }


class APIEventTypeNamespaceDetailView(generics.RetrieveAPIView):
    """
    get: Returns the details of an event type namespace.
    """
    serializer_class = EventTypeNamespaceSerializer

    def get_object(self):
        try:
            return EventTypeNamespace.get(
                name=self.kwargs['name']
            )
        except KeyError:
            raise Http404


class APIEventTypeNamespaceEventTypeListView(generics.ListAPIView):
    """
    get: Returns a list of all the available event types from a namespaces.
    """
    serializer_class = EventTypeSerializer

    def get_serializer_context(self):
        return {
            'format': self.format_kwarg,
            'request': self.request,
            'view': self
        }

    def get_source_queryset(self):
        try:
            return EventTypeNamespace.get(
                name=self.kwargs['name']
            ).get_event_types()
        except KeyError:
            raise Http404


class APIEventTypeNamespaceListView(generics.ListAPIView):
    """
    get: Returns a list of all the available event type namespaces.
    """
    serializer_class = EventTypeNamespaceSerializer
    source_queryset = EventTypeNamespace.all()

    def get_serializer_context(self):
        return {
            'format': self.format_kwarg,
            'request': self.request,
            'view': self
        }



from django.apps import apps
from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link

from .icons import (
    icon_document_file_all_signature_refresh,
    icon_document_file_all_signature_verify,
    icon_document_file_signature_detached_delete,
    icon_document_file_signature_detached_create,
    icon_document_file_signature_detail,
    icon_document_file_signature_detached_download,
    icon_document_file_signature_detached_upload,
    icon_document_file_signature_embedded_create,
    icon_document_file_signature_list
)
from .permissions import (
    permission_document_file_sign_detached,
    permission_document_file_sign_embedded,
    permission_document_file_signature_delete,
    permission_document_file_signature_download,
    permission_document_file_signature_upload,
    permission_document_file_signature_verify,
    permission_document_file_signature_view
)


def condition_is_detached_signature(context, resolved_object):
    SignatureBaseModel = apps.get_model(
        app_label='document_signatures', model_name='SignatureBaseModel'
    )

    return SignatureBaseModel.objects.select_subclasses().get(
        pk=context['resolved_object'].pk
    ).is_detached


# Detached signature

link_document_file_signature_detached_create = Link(
    args='resolved_object.pk',
    icon=icon_document_file_signature_detached_create,
    permission=permission_document_file_sign_detached,
    text=_(message='Sign detached'),
    view='signatures:document_file_signature_detached_create'
)
link_document_file_signature_detached_delete = Link(
    args='resolved_object.pk', condition=condition_is_detached_signature,
    icon=icon_document_file_signature_detached_delete,
    permission=permission_document_file_signature_delete,
    tags='dangerous', text=_(message='Delete'),
    view='signatures:document_file_signature_detached_delete'
)
link_document_file_signature_detached_download = Link(
    args='resolved_object.pk',
    icon=icon_document_file_signature_detached_download,
    condition=condition_is_detached_signature,
    permission=permission_document_file_signature_download,
    text=_(message='Download'),
    view='signatures:document_file_signature_detached_download'
)
link_document_file_signature_detached_upload = Link(
    args='resolved_object.pk',
    icon=icon_document_file_signature_detached_upload,
    permission=permission_document_file_signature_upload,
    text=_(message='Upload signature'),
    view='signatures:document_file_signature_detached_upload'
)

# Embedded

link_document_file_signature_embedded_create = Link(
    args='resolved_object.pk',
    icon=icon_document_file_signature_embedded_create,
    permission=permission_document_file_sign_embedded,
    text=_(message='Sign embedded'),
    view='signatures:document_file_signature_embedded_create'
)

# All

link_document_file_signature_detail = Link(
    args='resolved_object.pk',
    icon=icon_document_file_signature_detail,
    permission=permission_document_file_signature_view,
    text=_(message='Details'), view='signatures:document_file_signature_detail'
)

link_document_file_signature_list = Link(
    args='resolved_object.pk', icon=icon_document_file_signature_list,
    permission=permission_document_file_signature_view,
    text=_(message='Signatures'), view='signatures:document_file_signature_list'
)

# Tools

link_document_file_all_signature_refresh = Link(
    icon=icon_document_file_all_signature_refresh,
    permission=permission_document_file_signature_verify,
    text=_(message='Refresh all signatures'),
    view='signatures:all_document_file_signature_refresh'
)
link_document_file_all_signature_verify = Link(
    icon=icon_document_file_all_signature_verify,
    permission=permission_document_file_signature_verify,
    text=_(message='Verify all documents'),
    view='signatures:all_document_file_signature_verify'
)



from mayan.apps.documents.api_views.api_view_mixins import (
    ParentObjectDocumentFileAPIViewMixin
)
from mayan.apps.rest_api import generics

from ..permissions import permission_document_file_sources_metadata_view
from ..serializers import DocumentFileSourceMetadataSerializer


class APIDocumentFileSourceMetadataListView(
    ParentObjectDocumentFileAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of selected document's source metadata values.
    """
    serializer_class = DocumentFileSourceMetadataSerializer

    def get_source_queryset(self):
        document_file = self.get_document_file(
            permission=permission_document_file_sources_metadata_view
        )
        return document_file.source_metadata.all()


class APIDocumentFileSourceMetadataDetailView(
    ParentObjectDocumentFileAPIViewMixin, generics.RetrieveAPIView
):
    """
    get: Return the details of the selected document source metadata.
    """
    lookup_url_kwarg = 'document_file_source_metadata_id'
    serializer_class = DocumentFileSourceMetadataSerializer

    def get_source_queryset(self):
        document_file = self.get_document_file(
            permission=permission_document_file_sources_metadata_view
        )
        return document_file.source_metadata.all()



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_COMMON_DISABLE_LOCAL_STORAGE, DEFAULT_COMMON_DISABLED_APPS,
    DEFAULT_COMMON_EXTRA_APPS, DEFAULT_COMMON_EXTRA_APPS_PRE,
    DEFAULT_COMMON_HOME_VIEW, DEFAULT_COMMON_HOME_VIEW_DASHBOARD_NAME,
    DEFAULT_COMMON_PROJECT_TITLE
)

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Common'), name='common', version='0002'
)

setting_disable_local_storage = setting_namespace.do_setting_add(
    choices=('false', 'true'),
    default=DEFAULT_COMMON_DISABLE_LOCAL_STORAGE,
    global_name='COMMON_DISABLE_LOCAL_STORAGE', help_text=_(
        message='Disables the use of the local `media` folder. When using this '
        'setting, all apps must be also configured via their respective '
        'storage backend settings to use alternate persistence.'
    )
)
setting_disabled_apps = setting_namespace.do_setting_add(
    default=DEFAULT_COMMON_DISABLED_APPS,
    global_name='COMMON_DISABLED_APPS', help_text=_(
        message='A list of strings designating all applications that are to be '
        'removed from the list normally installed by Mayan EDMS. Each '
        'string should be a dotted Python path to: an application '
        'configuration class (preferred), or a package containing an '
        'application. Example: [\'app_1\', \'app_2\']'
    )
)
setting_extra_apps = setting_namespace.do_setting_add(
    default=DEFAULT_COMMON_EXTRA_APPS, global_name='COMMON_EXTRA_APPS',
    help_text=_(
        message='A list of strings designating all applications that are installed '
        'beyond those normally installed by Mayan EDMS. Each string '
        'should be a dotted Python path to: an application configuration '
        'class (preferred), or a package containing an application. '
        'These apps will be installed after the default apps. '
        'Example: [\'app_1\', \'app_2\']'
    )
)
setting_extra_apps_pre = setting_namespace.do_setting_add(
    default=DEFAULT_COMMON_EXTRA_APPS_PRE,
    global_name='COMMON_EXTRA_APPS_PRE',
    help_text=_(
        message='A list of strings designating all applications that are installed '
        'beyond those normally installed by Mayan EDMS. Each string '
        'should be a dotted Python path to: an application configuration '
        'class (preferred), or a package containing an application. '
        'These apps will be installed before the default apps. '
        'Example: [\'app_1\', \'app_2\']'
    )
)
setting_home_view = setting_namespace.do_setting_add(
    default=DEFAULT_COMMON_HOME_VIEW, global_name='COMMON_HOME_VIEW',
    help_text=_(
        message='Name of the view attached to the brand anchor in the main menu. '
        'This is also the view to which users will be redirected after '
        'log in.'
    )
)
setting_home_view_dashboard = setting_namespace.do_setting_add(
    default=DEFAULT_COMMON_HOME_VIEW_DASHBOARD_NAME,
    global_name='COMMON_HOME_VIEW_DASHBOARD_NAME',
    help_text=_(
        message='Name of the dashboard to display in the home view.'
    )
)
setting_project_title = setting_namespace.do_setting_add(
    default=DEFAULT_COMMON_PROJECT_TITLE, global_name='COMMON_PROJECT_TITLE',
    help_text=_(message='Sets the project\'s name.')
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link

from .icons import icon_model_property_list, icon_property_model_list


link_model_property_list = Link(
    icon=icon_model_property_list, kwargs={
        'app_label': 'resolved_object.app_config.label',
        'model_name': 'resolved_object.model_name',
    }, text=_(message='Properties'), view='databases:model_property_list'
)
link_property_model_list = Link(
    icon=icon_property_model_list, text=_(message='Property models'),
    view='databases:property_model_list'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_b, worker_c

from .literals import SOURCE_ACTION_EXECUTE_TASK_PATH

queue_sources = CeleryQueue(
    label=_(message='Sources'), name='sources', worker=worker_b
)
queue_sources_periodic = CeleryQueue(
    label=_(message='Sources periodic'), name='sources_periodic',
    transient=True,
    worker=worker_c
)

queue_sources.add_task_type(
    label=_(message='Handle source backend action background task'),
    dotted_path='mayan.apps.sources.tasks.task_source_backend_action_background_task'
)
queue_sources_periodic.add_task_type(
    label=_(message='Check interval source'),
    dotted_path=SOURCE_ACTION_EXECUTE_TASK_PATH
)



from django.utils.module_loading import import_string

from mayan.apps.mime_types.classes import MIMETypeBackend


class DownloadBackend:
    @classmethod
    def get_backend_class(cls, dotted_path):
        return import_string(dotted_path=dotted_path)

    @classmethod
    def get_backend_instance(cls, dotted_path, arguments=None):
        backend_class = cls.get_backend_class(dotted_path=dotted_path)

        kwargs = arguments or {}

        return backend_class(**kwargs)

    def get_download_filename(self, obj):
        return str(obj)

    def get_download_file_object(self, obj):
        return obj.open(mode='rb')

    def get_download_mime_type_and_encoding(self, obj):
        with obj.open(mode='rb') as file_object:
            mime_type, encoding = MIMETypeBackend.get_backend_instance().get_mime_type(
                file_object=file_object, mime_type_only=True
            )
            return (mime_type, encoding)



import logging

from django.utils.translation import gettext_lazy as _

from .exceptions import (
    DynamicSearchException, DynamicSearchQueryError,
    DynamicSearchValueTransformationError
)

logger = logging.getLogger(name=__name__)


class BackendQueryType:
    _registry = {}

    @classmethod
    def get_for_query_type(cls, search_backend, query_type):
        search_backend_dictionary = cls._registry.get(
            search_backend, {}
        )

        return search_backend_dictionary[query_type]

    @classmethod
    def register(cls, klass, search_backend):
        cls._registry.setdefault(
            search_backend, {}
        )

        search_backend_dictionary = cls._registry[search_backend]

        if klass.query_type in search_backend_dictionary:
            raise DynamicSearchException(
                'Backend query type `{}` is attempting to register to query '
                'type `{}` which is already registered to backend query '
                'type `{}`.'.format(
                    klass, klass.query_type,
                    search_backend_dictionary[klass.query_type]
                )
            )

        search_backend_dictionary[klass.query_type] = klass

    def __init__(
        self, is_quoted_value, search_backend, search_field, value,
        extra_kwargs=None
    ):
        self.extra_kwargs = extra_kwargs or {}
        self.is_quoted_value = is_quoted_value
        self.search_backend = search_backend
        self.search_field = search_field
        self.value = value

    def get_search_backend_field_type(self):
        return self.search_backend.get_search_field_backend_field_type(
            search_field=self.search_field
        )


class QueryTypeMetaclass(type):
    def __str__(cls):
        return str(cls.label)


class QueryType(metaclass=QueryTypeMetaclass):
    _default_klass = None
    _registry = []

    @classmethod
    def all(cls):
        return cls._registry

    @classmethod
    def check(cls, value):
        if value.startswith(cls.alias):
            return value[
                len(cls.alias):
            ]

    @classmethod
    def check_all(cls, value):
        for klass in cls.all():
            text = klass.check(value=value)
            if text is not None:
                return (klass, text)

        if cls._default_klass:
            return (cls._default_klass, value)
        else:
            raise DynamicSearchQueryError(
                'Query alias does not match any know query type class.'
            )

    @classmethod
    def do_value_process(
        cls, is_raw_value, search_backend, search_field, value
    ):
        if is_raw_value:
            return value
        else:
            return search_field.do_value_search_transform(
                search_backend=search_backend, value=value
            )

    @classmethod
    def get_default(cls):
        return cls._default_klass

    @classmethod
    def get_for_backend(cls, search_backend):
        return BackendQueryType.get_for_query_type(
            query_type=cls, search_backend=search_backend
        )

    @classmethod
    def resolve_for_backend(
        cls, is_quoted_value, is_raw_value, search_backend, search_field,
        value, extra_kwargs=None
    ):
        extra_kwargs = extra_kwargs or {}

        try:
            backend_query_type_class = cls.get_for_backend(
                search_backend=search_backend.__class__
            )
        except KeyError:
            raise DynamicSearchQueryError(
                'Query type `{}` is valid for field name `{}` but it is '
                'not implemented or registered by the backend.'.format(
                    cls, search_field.field_name
                )
            )
        else:
            if not value and not is_quoted_value:
                value = None

            try:
                value = cls.do_value_process(
                    is_raw_value=is_raw_value, search_backend=search_backend,
                    search_field=search_field, value=value
                )
            except DynamicSearchValueTransformationError as exception:
                logger.warning(
                    'Unable to convert value `{}` to the appropriate '
                    'data type to perform the search. Reason: {}'.format(
                        value, exception
                    )
                )
                raise
            else:
                backend_query_type = backend_query_type_class(
                    extra_kwargs=extra_kwargs,
                    is_quoted_value=is_quoted_value,
                    search_backend=search_backend, search_field=search_field,
                    value=value
                )

                return backend_query_type.do_resolve()

    @classmethod
    def register(cls, klass):
        cls._registry.append(klass)

    @classmethod
    def set_default(cls, klass):
        cls._default_klass = klass


class QueryTypeExact(QueryType):
    alias = '='
    label = _(message='Exact')
    name = 'EXACT'


class QueryTypeFuzzy(QueryType):
    alias = '~'
    label = _(message='Fuzzy')
    name = 'FUZZY'


class QueryTypeGreaterThan(QueryType):
    alias = '>'
    label = _(message='Greater than')
    name = 'GREATERTHAN'


class QueryTypeGreaterThanOrEqual(QueryType):
    alias = '>='
    label = _(message='Greater than or equal')
    name = 'GREATERTHANOREQUAL'


class QueryTypeLessThan(QueryType):
    alias = '<'
    label = _(message='Less than')
    name = 'LESSTHAN'


class QueryTypeLessThanOrEqual(QueryType):
    alias = '<='
    label = _(message='Less than or equal')
    name = 'LESSTHANOREQUAL'


class QueryTypePartial(QueryType):
    alias = '*'
    label = _(message='Partial')
    name = 'PARTIAL'


class QueryTypeRange(QueryType):
    alias = '[]'
    label = _(message='Range')
    name = 'RANGE'

    @classmethod
    def do_value_process(
        cls, is_raw_value, search_backend, search_field, value
    ):
        if is_raw_value:
            return value
        else:
            parts = []
            try:
                items = value.split('..')
            except AttributeError:
                raise DynamicSearchValueTransformationError(
                    'Unable to parse range term `{}`'.format(value)
                )
            else:
                if len(items) != 2:
                    raise DynamicSearchValueTransformationError(
                        'Invalid number of arguments for range `{}`'.format(
                            value
                        )
                    )
                for item in items:
                    parts.append(
                        search_field.do_value_search_transform(
                            search_backend=search_backend, value=item
                        )
                    )
                return parts


class QueryTypeRangeExclusive(QueryTypeRange):
    alias = '{}'
    label = _(message='Range exclusive')
    name = 'RANGEEXCLUSIVE'


class QueryTypeRegularExpression(QueryType):
    alias = '%'
    label = _(message='Regular expression')
    name = 'REGULAREXPRESSION'


QueryType.register(klass=QueryTypeExact)
QueryType.register(klass=QueryTypeFuzzy)
QueryType.register(klass=QueryTypeGreaterThanOrEqual)  # Must go before greater than.
QueryType.register(klass=QueryTypeGreaterThan)
QueryType.register(klass=QueryTypeLessThanOrEqual)  # Must go before less than.
QueryType.register(klass=QueryTypeLessThan)
QueryType.register(klass=QueryTypePartial)
QueryType.register(klass=QueryTypeRange)
QueryType.register(klass=QueryTypeRangeExclusive)
QueryType.register(klass=QueryTypeRegularExpression)

QueryType.set_default(klass=QueryTypePartial)



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('checkouts', '0010_auto_20200917_0643')
    ]

    operations = [
        migrations.DeleteModel(name='NewFileBlock')
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig


class FileMetadataEMLApp(MayanAppConfig):
    app_namespace = 'file_metadata_eml'
    app_url = 'file_metadata_eml'
    has_tests = True
    name = 'mayan.apps.file_metadata_eml'
    verbose_name = _(message='File metadata EML')



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .classes import DocumentFileCompressor
from .literals import (
    DEFAULT_DOCUMENT_FILE_DOWNLOAD_BACKEND,
    DEFAULT_DOCUMENT_FILE_DOWNLOAD_BACKEND_ARGUMENTS,
    DEFAULT_DOCUMENT_FILE_DOWNLOAD_MESSAGE_BODY,
    DEFAULT_DOCUMENT_FILE_DOWNLOAD_MESSAGE_SUBJECT
)

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Document downloads'), name='document_downloads'
)

setting_document_file_download_backend = setting_namespace.do_setting_add(
    default=DEFAULT_DOCUMENT_FILE_DOWNLOAD_BACKEND,
    global_name='DOCUMENT_DOWNLOADS_DOCUMENT_FILE_DOWNLOAD_BACKEND',
    help_text=_(
        'Path to the download subclass to use when downloading document '
        'files.'
    )
)
setting_document_file_download_backend_arguments = setting_namespace.do_setting_add(
    default=DEFAULT_DOCUMENT_FILE_DOWNLOAD_BACKEND_ARGUMENTS,
    global_name='DOCUMENT_DOWNLOADS_DOCUMENT_FILE_DOWNLOAD_BACKEND_ARGUMENTS',
    help_text=_(
        'Arguments to pass to the '
        'DOCUMENT_DOWNLOADS_DOCUMENT_FILE_DOWNLOAD_BACKEND.'
    )
)
setting_message_body_template = setting_namespace.do_setting_add(
    default=DEFAULT_DOCUMENT_FILE_DOWNLOAD_MESSAGE_BODY,
    global_name='DOCUMENT_FILE_DOWNLOAD_MESSAGE_BODY', help_text=_(
        'Template for the document download message body text. Can '
        'include HTML. Available variables: {}'.format(
            ', '.join(DocumentFileCompressor.context_key_list)
        )
    )
)
setting_message_subject_template = setting_namespace.do_setting_add(
    default=DEFAULT_DOCUMENT_FILE_DOWNLOAD_MESSAGE_SUBJECT,
    global_name='DOCUMENT_FILE_DOWNLOAD_MESSAGE_SUBJECT', help_text=_(
        'Template for the document download message subject line. '
        'Can\'t include HTML. Available variables: {}'.format(
            ', '.join(DocumentFileCompressor.context_key_list)
        )
    )
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.templating.template_backends import Template


class MixinConditionTemplate:
    def evaluate_condition(self, context=None):
        if self.has_condition():
            condition_context = context or self.get_condition_context()
            condition_template = self.get_condition_template()

            template = Template(template_string=condition_template)
            result = template.render(context=condition_context)
            result = result.strip()

            return result
        else:
            return True

    def get_condition_context(self):
        return {}

    def get_condition_template(self):
        return self.condition

    def has_condition(self):
        condition_template = self.get_condition_template()

        if condition_template:
            return True
        else:
            return False
    has_condition.help_text = _(
        message='The object will be executed, depending on the condition '
        'return value.'
    )
    has_condition.short_description = _(message='Has a condition?')



from django.utils.functional import classproperty

from ..exceptions import (
    SourceActionException, SourceActionExceptionInterface,
    SourceActionExceptionInterfaceUnknown
)
from ..tasks import task_source_backend_action_background_task

from .interfaces import SourceBackendActionInterface
from .mixins.source_metadata_mixins import (
    SourceBackendActionMixinSourceMetadata
)


class SourceBackendActionMetaclass(type):
    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(
            mcs, name, bases, attrs
        )

        interface_dict = {}
        for mro_klass in new_class.mro():
            interface_base = getattr(mro_klass, 'Interface', None)
            if interface_base:
                for item_name in dir(interface_base):
                    item = getattr(interface_base, item_name)
                    try:
                        if issubclass(item, SourceBackendActionInterface):
                            interface_dict.setdefault(
                                item_name, []
                            )
                            if item not in interface_dict[item_name]:
                                interface_dict[item_name].append(item)
                    except TypeError:
                        pass

        if interface_dict:
            class InterfaceNamespace:
                """
                Temporary class to hold all computed interface subclasses.
                """

            for key, value in interface_dict.items():
                InterfaceClass = type(
                    key, tuple(value), {}
                )

                setattr(InterfaceNamespace, key, InterfaceClass)

            setattr(new_class, 'Interface', InterfaceNamespace)

        return new_class


class SourceBackendActionBase(metaclass=SourceBackendActionMetaclass):
    accept_files = False
    confirmation = True  # Require a submit or a POST to execute.
    name = None
    permission = None

    class Interface:
        """
        Class which all actions are expected to have.
        """

    @classproperty
    def interface_visible_list(cls):
        # Used by the REST API serializer.
        for interface in cls.get_interface_list():
            if not interface.hidden:
                yield interface

    @classmethod
    def get_argument_list(cls, interface_name):
        try:
            action_interface = cls.get_interface_class(
                interface_name=interface_name
            )
        except AttributeError:
            return ()
        else:
            return action_interface.get_arguments()

    @classmethod
    def get_interface_list(cls):
        action_interface_namespace = getattr(cls, 'Interface', None)
        interface_list = []

        if action_interface_namespace:
            for item_name in dir(action_interface_namespace):
                item = getattr(action_interface_namespace, item_name)
                try:
                    if issubclass(item, SourceBackendActionInterface):
                        interface_list.append(item)
                except TypeError:
                    pass

        return interface_list

    def __init__(self, source):
        self.source = source

    def __str__(self):
        return 'Source backend action "{}"'.format(self.name)

    def _background_task(self, **kwargs):
        return {}

    def _execute(self, **kwargs):
        background_task_kwargs = self.get_task_kwargs(**kwargs)

        task_source_backend_action_background_task.apply_async(
            kwargs=background_task_kwargs
        )

    def background_task(self, interface_load_kwargs=None):
        interface_load_kwargs = interface_load_kwargs or {}

        interface_instance = self.get_interface_instance(
            interface_name='Task'
        )

        task_kwargs = interface_instance.load(**interface_load_kwargs)

        self._background_task(**task_kwargs)

    def execute(
        self, interface_name, interface_load_kwargs=None,
        interface_retrieve_kwargs=None
    ):
        interface_load_kwargs = interface_load_kwargs or {}
        interface_retrieve_kwargs = interface_retrieve_kwargs or {}
        interface_instance = self.get_interface_instance(
            interface_name=interface_name
        )

        try:
            # Only intercept action exception to add additional content.
            # All other should be ignored. They might exception with
            # interface context like Http404 for views or REST API.
            execute_kwargs = interface_instance.load(**interface_load_kwargs)
        except SourceActionException as exception:
            raise SourceActionException(
                'Unable to execute action `{}`; {}'.format(
                    self.name, exception
                )
            ) from exception

        backend_instance = self.source.get_backend_instance()

        allow_action_execution = backend_instance.get_allow_action_execute(
            action=self, action_execute_kwargs=execute_kwargs
        )

        if allow_action_execution:
            action_data = self._execute(**execute_kwargs)
        else:
            action_data = None

        return interface_instance.retrieve(
            context=interface_retrieve_kwargs, data=action_data
        )

    def get_confirmation_context(
        self, interface_name, interface_load_kwargs=None
    ):
        interface_load_kwargs = interface_load_kwargs or {}

        interface_instance = self.get_interface_instance(
            interface_name=interface_name
        )

        try:
            result = interface_instance.get_confirmation_context(
                **interface_load_kwargs
            )
        except Exception as exception:
            raise SourceActionExceptionInterface(
                'Unable to get interface confirmation context; {}'.format(
                    exception
                )
            ) from exception

        return result

    def get_document_file_task_kwargs(self, **kwargs):
        return {}

    def get_document_task_kwargs(self, **kwargs):
        return {}

    def get_interface_class(self, interface_name):
        action_interface_namespace = getattr(self, 'Interface')

        try:
            return getattr(action_interface_namespace, interface_name)
        except AttributeError:
            raise SourceActionExceptionInterfaceUnknown(
                'Unknown interface `{}` for action `{}`.'.format(
                    interface_name, self.name
                )
            )

    def get_interface_instance(self, interface_name):
        action_interface_class = self.get_interface_class(
            interface_name=interface_name
        )
        return action_interface_class(action=self)

    def get_task_kwargs(self):
        return {
            'action_name': self.name,
            'action_interface_kwargs': {},
            'source_id': self.source.pk
        }

    def has_interface(self, interface_name):
        return hasattr(self.Interface, interface_name)


class SourceBackendAction(
    SourceBackendActionMixinSourceMetadata, SourceBackendActionBase
):
    """
    Class to define actions per source backends. Each action also defines
    interfaces. The action backend will process the input data and transform
    the source backend output for each interface class.
    """


class SourceBackendActionDummy(SourceBackendAction):
    """
    Blank source action class for use in API documentation.
    """



from django.utils.translation import gettext_lazy as _

from mayan.apps.forms import form_widgets, forms

from .literals import STATE_LABELS
from .models import DocumentCheckout
from .widgets import SplitTimeDeltaWidget


class DocumentCheckOutForm(forms.ModelForm):
    class Meta:
        fields = ('expiration_datetime', 'block_new_file')
        model = DocumentCheckout
        widgets = {
            'expiration_datetime': SplitTimeDeltaWidget()
        }


class DocumentCheckOutDetailForm(forms.DetailForm):
    def __init__(self, *args, **kwargs):
        instance = kwargs['instance']

        extra_fields = (
            {
                'label': _(message='Document status'),
                'func': lambda instance: STATE_LABELS[
                    instance.get_check_out_state()
                ]
            },
        )

        if instance.is_checked_out():
            checkout_info = instance.get_check_out_info()
            extra_fields += (
                {
                    'label': _(message='User'),
                    'func': lambda instance: checkout_info.user.get_full_name() or checkout_info.user
                },
                {
                    'label': _(message='Check out time'),
                    'func': lambda instance: checkout_info.checkout_datetime,
                    'widget': form_widgets.DateTimeInput
                },
                {
                    'label': _(message='Check out expiration'),
                    'func': lambda instance: checkout_info.expiration_datetime,
                    'widget': form_widgets.DateTimeInput
                },
                {
                    'label': _(message='New files allowed?'),
                    'func': lambda instance: _(message='Yes') if not checkout_info.block_new_file else _(message='No')
                }
            )

        kwargs['extra_fields'] = extra_fields
        super().__init__(*args, **kwargs)

    class Meta:
        fields = ()
        model = DocumentCheckout



from django.http import Http404

from django.utils.translation import gettext_lazy as _

from mayan.apps.views.generics import SingleObjectListView

from ..icons import icon_search_model_detail, icon_search_model_list
from ..search_models import SearchModel


class SearchModelListView(SingleObjectListView):
    extra_context = {
        'hide_link': True,
        'hide_object': True,
        'title': _(message='Search models')
    }
    view_icon = icon_search_model_list

    def get_source_queryset(self):
        return SearchModel.all()


class SearchModelSearchFieldListView(SingleObjectListView):
    view_icon = icon_search_model_detail

    def get_extra_context(self):
        search_model = self.get_search_model()

        return {
            'hide_link': True,
            'hide_object': True,
            'object': search_model,
            'title': _(
                message='Fields for search model: %s'
            ) % search_model.label
        }

    def get_search_model(self):
        try:
            return SearchModel.get(
                name=self.kwargs['search_model_name']
            )
        except KeyError:
            raise Http404(
                _(message='Search model: %s, not found') % self.kwargs[
                    'search_model_name'
                ]
            )

    def get_source_queryset(self):
        search_model = self.get_search_model()
        return search_model.search_fields_label_sorted



from pypdf import PdfReader

from django.utils.translation import gettext_lazy as _

from mayan.apps.file_metadata.classes import FileMetadataDriver


class FileMetadataDriverPyPDF(FileMetadataDriver):
    description = _(message='Read meta information stored in files.')
    internal_name = 'pypdf'
    label = _(message='PyPDF')
    mime_type_list = ('application/pdf',)

    def _process(self, document_file):
        with PdfReader(stream=document_file.file) as reader:
            return reader.metadata



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = []

    operations = [
        migrations.CreateModel(
            name='Key',
            fields=[
                (
                    'id', models.AutoField(
                        verbose_name='ID', serialize=False,
                        auto_created=True, primary_key=True
                    )
                ),
                (
                    'data', models.TextField(verbose_name='Data')
                ),
                (
                    'key_id', models.CharField(
                        unique=True, max_length=16, verbose_name='Key ID'
                    )
                ),
                (
                    'creation_date', models.DateField(
                        verbose_name='Creation date'
                    )
                ),
                (
                    'expiration_date', models.DateField(
                        null=True, verbose_name='Expiration date', blank=True
                    )
                ),
                (
                    'fingerprint', models.CharField(
                        unique=True, max_length=40, verbose_name='Fingerprint'
                    )
                ),
                (
                    'length', models.PositiveIntegerField(verbose_name='Length')
                ),
                (
                    'algorithm', models.PositiveIntegerField(
                        verbose_name='Algorithm'
                    )
                ),
                (
                    'user_id', models.TextField(
                        verbose_name='User ID'
                    )
                ),
                (
                    'key_type', models.CharField(
                        max_length=3, verbose_name='Type'
                    )
                )
            ],
            options={
                'verbose_name': 'Key',
                'verbose_name_plural': 'Keys'
            }
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Templating'), name='templating'
)

permission_template_sandbox = namespace.add_permission(
    label=_(message='Use the template sandbox'), name='template_sandbox'
)



from mayan.apps.dependencies.classes import PythonDependency

PythonDependency(
    module=__name__, name='gevent', version_string='==25.8.2'
)
PythonDependency(
    module=__name__, name='greenlet', version_string='==3.2.4'
)
PythonDependency(
    module=__name__, name='gunicorn', version_string='==23.0.0'
)
PythonDependency(
    module=__name__, name='whitenoise', version_string='==6.9.0'
)



from django.apps import apps
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.acls.permissions import (
    permission_acl_edit, permission_acl_view
)
from mayan.apps.app_manager.apps import MayanAppConfig
from mayan.apps.common.classes import ModelCopy
from mayan.apps.common.menus import (
    menu_multi_item, menu_object, menu_return, menu_secondary, menu_setup
)
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.forms import column_widgets
from mayan.apps.logging.classes import ErrorLog, ErrorLogDomain
from mayan.apps.navigation.source_columns import SourceColumn

from .classes import MailerBackend
from .events import event_email_sent, event_mailing_profile_edited
from .links import (
    link_document_file_attachment_send_multiple,
    link_document_file_attachment_send_single,
    link_document_file_link_send_multiple,
    link_document_file_link_send_single, link_document_link_send_multiple,
    link_document_link_send_single,
    link_document_version_attachment_send_multiple,
    link_document_version_attachment_send_single,
    link_document_version_link_send_multiple,
    link_document_version_link_send_single,
    link_mailing_profile_create, link_mailing_profile_delete,
    link_mailing_profile_edit, link_mailing_profile_list,
    link_mailing_profile_setup, link_mailing_profile_test
)
from .literals import (
    DOCUMENT_FILE_CONTENT_FUNCTION_DOTTED_PATH,
    DOCUMENT_FILE_MIME_TYPE_FUNCTION_DOTTED_PATH,
    DOCUMENT_VERSION_CONTENT_FUNCTION_DOTTED_PATH,
    DOCUMENT_VERSION_MIME_TYPE_FUNCTION_DOTTED_PATH, ERROR_LOG_DOMAIN_NAME
)
from .mailing_actions import (
    ModelMailingActionAttachment, ModelMailingActionLink
)
from .permissions import (
    permission_send_document_file_attachment,
    permission_send_document_file_link, permission_send_document_link,
    permission_send_document_version_attachment,
    permission_send_document_version_link, permission_mailing_profile_delete,
    permission_mailing_profile_edit, permission_mailing_profile_use,
    permission_mailing_profile_view
)


class MailerApp(MayanAppConfig):
    app_namespace = 'mailer'
    app_url = 'mailer'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.mailer'
    verbose_name = _(message='Mailer')

    def ready(self):
        super().ready()

        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        DocumentFile = apps.get_model(
            app_label='documents', model_name='DocumentFile'
        )
        DocumentVersion = apps.get_model(
            app_label='documents', model_name='DocumentVersion'
        )

        UserMailer = self.get_model(model_name='UserMailer')

        error_log = ErrorLog(app_config=self)
        error_log.register_model(model=UserMailer, register_permission=True)

        ErrorLogDomain(
            label=_(message='Mailer'), name=ERROR_LOG_DOMAIN_NAME
        )

        EventModelRegistry.register(model=UserMailer)

        MailerBackend.load_modules()

        ModelCopy(
            model=UserMailer, bind_link=True, register_permission=True
        ).add_fields(
            field_names=(
                'label', 'enabled', 'backend_path', 'backend_data'
            ), field_values={
                'default': False
            }
        )

        ModelEventType.register(
            model=UserMailer, event_types=(
                event_email_sent, event_mailing_profile_edited
            )
        )

        ModelMailingActionLink(
            manager_name='valid', model=Document,
            permission=permission_send_document_link
        )

        ModelMailingActionAttachment(
            content_function_dotted_path=DOCUMENT_FILE_CONTENT_FUNCTION_DOTTED_PATH,
            manager_name='valid',
            mime_type_function_dotted_path=DOCUMENT_FILE_MIME_TYPE_FUNCTION_DOTTED_PATH,
            model=DocumentFile,
            permission=permission_send_document_file_attachment
        )
        ModelMailingActionLink(
            manager_name='valid',
            model=DocumentFile,
            permission=permission_send_document_file_link
        )

        ModelMailingActionAttachment(
            content_function_dotted_path=DOCUMENT_VERSION_CONTENT_FUNCTION_DOTTED_PATH,
            manager_name='valid',
            mime_type_function_dotted_path=DOCUMENT_VERSION_MIME_TYPE_FUNCTION_DOTTED_PATH,
            model=DocumentVersion,
            permission=permission_send_document_version_attachment
        )
        ModelMailingActionLink(
            manager_name='valid',
            model=DocumentVersion,
            permission=permission_send_document_version_link
        )

        ModelPermission.register(
            model=UserMailer, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_mailing_profile_delete,
                permission_mailing_profile_edit,
                permission_mailing_profile_use,
                permission_mailing_profile_view
            )
        )

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=UserMailer
        )
        SourceColumn(
            attribute='default', include_label=True, is_sortable=True,
            source=UserMailer, widget=column_widgets.TwoStateWidget
        )
        SourceColumn(
            attribute='enabled', include_label=True, is_sortable=True,
            source=UserMailer, widget=column_widgets.TwoStateWidget
        )
        SourceColumn(
            attribute='get_backend_class_label', include_label=True,
            source=UserMailer
        )

        # Document

        menu_multi_item.bind_links(
            links=(
                link_document_link_send_multiple,
            ), sources=(Document,)
        )

        menu_object.bind_links(
            links=(
                link_document_link_send_single,
            ), sources=(Document,)
        )

        # Document file

        menu_multi_item.bind_links(
            links=(
                link_document_file_attachment_send_multiple,
                link_document_file_link_send_multiple
            ), sources=(DocumentFile,)
        )

        menu_object.bind_links(
            links=(
                link_document_file_link_send_single,
                link_document_file_attachment_send_single
            ), sources=(DocumentFile,)
        )

        # Document version

        menu_multi_item.bind_links(
            links=(
                link_document_version_attachment_send_multiple,
                link_document_version_link_send_multiple
            ), sources=(DocumentVersion,)
        )

        menu_object.bind_links(
            links=(
                link_document_version_link_send_single,
                link_document_version_attachment_send_single
            ), sources=(DocumentVersion,)
        )

        # Mailing profile

        menu_object.bind_links(
            links=(
                link_mailing_profile_delete, link_mailing_profile_edit,
                link_mailing_profile_test
            ), sources=(UserMailer,)
        )

        menu_return.bind_links(
            links=(link_mailing_profile_list,), sources=(
                UserMailer, 'mailer:mailing_profile_list',
                'mailer:mailing_profile_create'
            )
        )

        menu_secondary.bind_links(
            links=(link_mailing_profile_create,), sources=(
                UserMailer, 'mailer:mailing_profile_list',
                'mailer:mailing_profile_create'
            )
        )

        menu_setup.bind_links(
            links=(link_mailing_profile_setup,)
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.dependencies.classes import JavaScriptDependency

JavaScriptDependency(
    label=_(message='FontAwesome'), module=__name__,
    name='@fortawesome/fontawesome-free', version_string='=6.7.2'
)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('locales', '0010_alter_userlocaleprofile_timezone')
    ]

    operations = [
        migrations.AlterField(
            model_name='userlocaleprofile', name='language',
            field=models.CharField(
                choices=[
                    ('sq', 'Albanian'), ('ar', 'Arabic'),
                    ('ar-eg', 'Arabic (Egypt)'),
                    ('hy-am', 'Armenian (Armenia)'), ('bs', 'Bosnian'),
                    ('bg', 'Bulgarian'), ('ca', 'Catalan'),
                    ('zh-cn', 'Chinese (China)'),
                    ('zh-hans', 'Chinese (Simplified)'),
                    ('zh-tw', 'Chinese (Taiwan)'), ('hr', 'Croatian'),
                    ('cs', 'Czech'), ('da', 'Danish'), ('nl', 'Dutch'),
                    ('en', 'English'), ('fr', 'French'),
                    ('fr-fr', 'French (France)'),
                    ('de-at', 'German (Austria)'),
                    ('de-de', 'German (Germany)'), ('el', 'Greek'),
                    ('he-il', 'Hebrew (Israel)'), ('hu', 'Hungarian'),
                    ('hu-hu', 'Hungarian (Hungary)'),
                    ('hu-sk', 'Hungarian (Slovakia)'), ('id', 'Indonesian'),
                    ('it', 'Italian'), ('lv', 'Latvian'),
                    ('mn-mn', 'Mongolian (Mongolia)'), ('fa', 'Persian'),
                    ('fa-ir', 'Persian (Iran)'), ('pl', 'Polish'),
                    ('pt', 'Portuguese'), ('pt-br', 'Portuguese (Brazil)'),
                    ('ro-ro', 'Romanian (Romania)'), ('ru', 'Russian'),
                    ('sl', 'Slovenian'), ('es', 'Spanish'),
                    ('es-ec', 'Spanish (Ecuador)'),
                    ('es-mx', 'Spanish (Mexico)'),
                    ('es-pr', 'Spanish (Puerto Rico)'), ('th', 'Thai'),
                    ('bo', 'Tibetan'), ('tr', 'Turkish'),
                    ('tr-tr', 'Turkish (Turkey)'), ('uk', 'Ukrainian'),
                    ('vi', 'Vietnamese')
                ], max_length=8, verbose_name='Language'
            )
        )
    ]



from django.utils.translation import gettext_lazy as _

DEFAULT_DOCUMENT_BODY_TEMPLATE = _(
    message='Attached to this email is the {{ object_name }}: {{ object }}\n\n '
    '--------\n '
    'This email has been sent from %(project_title)s (%(project_website)s)'
)
DEFAULT_DOCUMENT_SUBJECT_TEMPLATE = _(
    message='{{ object_name }}: {{ object }}'
)
DEFAULT_LINK_BODY_TEMPLATE = _(
    message='To access this {{ object_name }} click on the following link: '
    '{{ link }}\n\n--------\n '
    'This email has been sent from %(project_title)s (%(project_website)s)'
)
DEFAULT_LINK_SUBJECT_TEMPLATE = _(
    message='Link for {{ object_name }}: {{ object }}'
)
EMAIL_SEPARATORS = (',', ';')

ERROR_LOG_DOMAIN_NAME = 'mailer'

DOCUMENT_FILE_CONTENT_FUNCTION_DOTTED_PATH = 'mayan.apps.mailer.utils.get_document_file_content'
DOCUMENT_FILE_MIME_TYPE_FUNCTION_DOTTED_PATH = 'mayan.apps.mailer.utils.get_document_file_mime_type'

DOCUMENT_VERSION_CONTENT_FUNCTION_DOTTED_PATH = 'mayan.apps.mailer.utils.get_document_version_content'
DOCUMENT_VERSION_MIME_TYPE_FUNCTION_DOTTED_PATH = 'mayan.apps.mailer.utils.get_document_version_mime_type'



from functools import cache
import logging
from pathlib import Path
import sys
import traceback

from django.apps import apps, AppConfig
from django.urls import include, re_path
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.menus import menu_tools
from mayan.apps.common.utils import get_class_full_name
from mayan.apps.forms import column_widgets
from mayan.apps.navigation.source_columns import SourceColumn
from mayan.apps.organizations.settings import (
    setting_organization_url_base_path
)
from mayan.settings import BASE_DIR

from .links import link_app_list

logger = logging.getLogger(name=__name__)


class MayanAppConfig(AppConfig):
    app_namespace = None
    app_url = None
    has_app_translations = True
    has_rest_api = False
    has_tests = False

    @classmethod
    def get_all(cls):
        list_apps = [
            app for app in apps.get_app_configs() if issubclass(app.__class__, MayanAppConfig)
        ]

        return sorted(
            list_apps, key=lambda x: str(x.verbose_name)
        )

    @classmethod
    @cache
    def get_by_namespace(cls, namespace):
        for app_config in apps.get_app_configs():
            app_namespace = getattr(app_config, 'app_namespace', None)

            if app_namespace == namespace:
                return app_config

        raise KeyError(
            'No app found with `namespace` "{}"'.format(namespace)
        )

    def configure_urls(self):
        # Hidden import.
        from mayan.urls import urlpatterns as mayan_urlpatterns

        installation_base_url = setting_organization_url_base_path.value
        if installation_base_url:
            installation_base_url = '{}/'.format(installation_base_url)
        else:
            installation_base_url = ''

        if self.app_url:
            top_url = '{installation_base_url}{app_urls}/'.format(
                app_urls=self.app_url,
                installation_base_url=installation_base_url
            )
        elif self.app_url is not None:
            # When using app_url as '' to register a top of URL view.
            top_url = installation_base_url
        else:
            # If app_url is None, use the app's name for the URL base.
            top_url = '{installation_base_url}{app_name}/'.format(
                app_name=self.name,
                installation_base_url=installation_base_url
            )

        try:
            app_urlpatterns = import_string(
                dotted_path='{}.urls.urlpatterns'.format(self.name)
            )
        except ImportError as exception:
            non_critical_error_list = (
                'No module named urls',
                'No module named \'{}.urls\''.format(self.name),
                'Module "{}.urls" does not define a "urlpatterns" attribute/class'.format(self.name)
            )
            if str(exception) not in non_critical_error_list:
                logger.exception(
                    'Import time error when running AppConfig.ready() of app '
                    '"%s".', self.name
                )
                exc_info = sys.exc_info()
                traceback.print_exception(*exc_info)
                raise exception
        else:
            # Allow blank namespaces. These are used to register the
            # urlpatterns of encapsulated libraries as top level named
            # URLs.
            if self.app_namespace is not None:
                app_namespace = self.app_namespace
            else:
                app_namespace = self.name

            mayan_urlpatterns += (
                re_path(
                    route=r'^{}'.format(top_url), view=include(
                        (app_urlpatterns, app_namespace)
                    )
                ),
            )

        try:
            passthru_urlpatterns = import_string(
                dotted_path='{}.urls.passthru_urlpatterns'.format(self.name)
            )
        except ImportError as exception:
            non_critical_error_list = (
                'No module named urls',
                'No module named \'{}.urls\''.format(self.name),
                'Module "{}.urls" does not define a "passthru_urlpatterns" attribute/class'.format(self.name)
            )
            if str(exception) not in non_critical_error_list:
                logger.exception(
                    'Import time error when running AppConfig.ready() of app '
                    '"%s".', self.name
                )
                exc_info = sys.exc_info()
                traceback.print_exception(*exc_info)
                raise exception
        else:
            mayan_urlpatterns += (
                re_path(
                    route=r'^{}'.format(top_url), view=include(
                        passthru_urlpatterns
                    )
                ),
            )

    def get_app_class_full_path(self):
        return get_class_full_name(klass=self.__class__)

    get_app_class_full_path.short_description = _(
        message='App class full path'
    )

    def get_app_url(self):
        return self.app_url or ''

    get_app_url.short_description = _(message='App URL')

    def get_has_app_translations(self):
        return getattr(self, 'has_app_translations', True)

    def get_has_javascript_translations(self):
        return getattr(self, 'has_javascript_translations', False)

    def get_has_rest_api(self):
        return getattr(self, 'has_rest_api', False)

    get_has_rest_api.short_description = _(message='Has REST API')

    def get_has_tests(self):
        return getattr(self, 'has_tests', False)

    get_has_tests.short_description = _(message='Has tests')

    def get_name(self):
        return self.name

    get_name.short_description = _(message='Name')

    def get_verbose_name(self):
        return self.verbose_name

    get_verbose_name.short_description = _(message='Verbose name')

    @property
    def mayan_path(self):
        return Path(self.path)

    @property
    def mayan_path_relative(self):
        return self.mayan_path.relative_to(BASE_DIR.parent)

    def ready(self):
        logger.debug('Initializing app: %s', self.name)
        self.configure_urls()


class AppManagerAppConfig(MayanAppConfig):
    app_namespace = 'app_manager'
    app_url = 'app_manager'
    name = 'mayan.apps.app_manager'
    verbose_name = _(message='App manager')

    def ready(self):
        super().ready()

        SourceColumn(
            attribute='get_verbose_name', include_label=True,
            is_identifier=True, source=MayanAppConfig
        )
        SourceColumn(
            attribute='get_app_url', include_label=True,
            source=MayanAppConfig
        )
        SourceColumn(
            attribute='get_name', include_label=True,
            source=MayanAppConfig
        )
        SourceColumn(
            attribute='get_app_class_full_path', include_label=True,
            source=MayanAppConfig
        )
        SourceColumn(
            attribute='get_has_rest_api', include_label=True, source=MayanAppConfig,
            widget=column_widgets.TwoStateWidget
        )
        SourceColumn(
            attribute='get_has_tests', include_label=True, source=MayanAppConfig,
            widget=column_widgets.TwoStateWidget
        )

        menu_tools.bind_links(
            links=(link_app_list,)
        )



import logging

from furl import furl

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.template import RequestContext, Variable, VariableDoesNotExist
from django.template.defaulttags import URLNode
from django.urls import resolve, reverse
from django.utils.encoding import force_str

from mayan.apps.common.settings import setting_home_view
from mayan.apps.permissions.classes import Permission

from .class_mixins import TemplateObjectMixin

logger = logging.getLogger(name=__name__)


class Link(TemplateObjectMixin):
    _registry = {}

    @staticmethod
    def conditional_active_by_view_name(context, resolved_link):
        return resolved_link.link.view == resolved_link.current_view_name

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def remove(cls, name):
        del cls._registry[name]

    def __init__(
        self, text=None, view=None, args=None, badge_text=None,
        condition=None, conditional_active=None, conditional_disable=None,
        description=None, html_data=None, html_extra_classes=None, icon=None,
        keep_query=False, kwargs=None, name=None, permission=None, query=None,
        remove_from_query=None, tags=None, title=None, url=None
    ):
        self.args = args or []
        self.badge_text = badge_text
        self.condition = condition
        self.conditional_active = conditional_active or Link.conditional_active_by_view_name
        self.conditional_disable = conditional_disable
        self.description = description
        self.html_data = html_data
        self.html_data_resolved = None
        self.html_extra_classes = html_extra_classes
        self._icon = icon
        self.keep_query = keep_query
        self._kwargs = kwargs or {}
        self.name = name
        self._permission = permission
        self.query = query or {}
        self.remove_from_query = remove_from_query or []
        self.tags = tags
        self.text = text
        self.title = title
        self.view = view
        self.url = url

        if name:
            self.__class__._registry[name] = self

    def get_icon(self, context=None):
        return self._icon

    def get_kwargs(self, context):
        try:
            return self._kwargs(context)
        except TypeError:
            # Is not a callable.
            return self._kwargs

    def get_permission_object(self, context):
        return None

    def get_permission(self, context):
        return self._permission

    def get_resolved_object(self, context):
        try:
            return Variable(
                var='object'
            ).resolve(context=context)
        except VariableDoesNotExist:
            """No object variable in the context"""

    def resolve(self, context=None, request=None, resolved_object=None):
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )

        if not context and not request:
            raise ImproperlyConfigured(
                'Must provide a context or a request in order to resolve the '
                'link.'
            )

        if not context:
            context = RequestContext(request=request)

        request = self.get_request(context=context, request=request)

        current_path = request.META['PATH_INFO']
        current_view_name = resolve(path=current_path).view_name

        if not resolved_object:
            resolved_object = self.get_resolved_object(context=context)

        # If we were passed an instance of the view context object we are
        # resolving, inject it into the context. This helps resolve links for
        # object lists.
        if resolved_object:
            context['resolved_object'] = resolved_object

        # ACL is tested against the resolved_object, {{ object }}
        # or a custom object returned by the link subclass.
        permission_object = self.get_permission_object(context=context) or resolved_object

        # If this link has a required permission check that the user has it
        # too.
        permission = self.get_permission(context=context)

        if permission:
            if permission_object:
                try:
                    AccessControlList.objects.check_access(
                        obj=permission_object, permission=permission,
                        user=request.user
                    )
                except PermissionDenied:
                    return None
            else:
                try:
                    Permission.check_user_permission(
                        permission=permission, user=request.user
                    )
                except PermissionDenied:
                    return None

        # Check to see if link has conditional display function and only
        # display it if the result of the conditional display function is
        # True.
        if not self.check_condition(context=context, resolved_object=resolved_object):
            return None

        resolved_link = ResolvedLink(
            current_view_name=current_view_name, link=self
        )

        if self.view:
            view_name = Variable(
                var='"{}"'.format(self.view)
            )
            if isinstance(self.args, list) or isinstance(self.args, tuple):
                args = [
                    Variable(var=arg) for arg in self.args
                ]
            else:
                args = [
                    Variable(var=self.args)
                ]

            kwargs = self.get_kwargs(context=context)

            kwargs = {
                key: Variable(var=value) for key, value in kwargs.items()
            }

            # Use Django's exact {% url %} code to resolve the link.
            node = URLNode(
                view_name=view_name, args=args, kwargs=kwargs, asvar=None
            )
            try:
                resolved_link.url = node.render(context=context)
            except VariableDoesNotExist as exception:
                """Not critical, ignore"""
                logger.debug(
                    'VariableDoesNotExist when resolving link "%s" '
                    'URL; %s', self.text, exception
                )
            except Exception as exception:
                logger.error(
                    'Error resolving link "%s" URL; %s', self.text,
                    exception, exc_info=True
                )
        elif self.url is not None:
            resolved_link.url = self.url

        if self.html_data:
            result = {}
            for key, value in self.html_data.items():
                try:
                    resolved_value = Variable(
                        var=value
                    ).resolve(context=context)
                except VariableDoesNotExist:
                    """No object variable in the context"""
                    resolved_value = value

                result[key] = resolved_value

            self.html_data_resolved = result

        # This is for links that should be displayed but that are not
        # clickable.
        if self.conditional_disable:
            resolved_link.disabled = self.conditional_disable(
                context=context
            )
        else:
            resolved_link.disabled = False

        # Lets a new link keep the same URL query string of the current URL.
        if self.keep_query:
            # Sometimes we are required to remove a key from the URL query.
            parsed_url = furl(
                force_str(
                    s=request.get_full_path() or request.META.get(
                        'HTTP_REFERER', reverse(setting_home_view.value)
                    )
                )
            )

            for key in self.remove_from_query:
                try:
                    parsed_url.query.remove(key)
                except KeyError:
                    pass

            # Use the link's URL but with the previous URL querystring.
            new_url = furl(url=resolved_link.url)
            new_url.args = parsed_url.querystr
            resolved_link.url = new_url.url

        if self.query:
            new_url = furl(url=resolved_link.url)
            for key, value in self.query.items():
                try:
                    resolved_variable = Variable(
                        var=value
                    ).resolve(context=context)
                except VariableDoesNotExist:
                    """
                    Not fatal. Variable resolution here is perform as if
                    resolving a template variable. Non existing results
                    are not updated.
                    """
                else:
                    new_url.args[key] = resolved_variable

            resolved_link.url = new_url.url

        resolved_link.context = context

        return resolved_link


class ResolvedLink:
    def __init__(self, link, current_view_name):
        self.context = None
        self.current_view_name = current_view_name
        self.disabled = False
        self.link = link
        self.request = None
        self.url = '#'

    def __repr__(self):
        return '<ResolvedLink: {}>'.format(self.url)

    @property
    def active(self):
        conditional_active = self.link.conditional_active
        if conditional_active:
            return conditional_active(
                context=self.context, resolved_link=self
            )

    @property
    def badge_text(self):
        if self.link.badge_text:
            return self.link.badge_text(context=self.context)

    @property
    def description(self):
        return self.link.description

    def get_icon(self, context=None):
        return self.link.get_icon(context=context)

    @property
    def html_data(self):
        return self.link.html_data_resolved

    @property
    def html_extra_classes(self):
        return self.link.html_extra_classes or ''

    @property
    def tags(self):
        return self.link.tags

    @property
    def text(self):
        try:
            return self.link.text(context=self.context)
        except TypeError:
            return self.link.text

    @property
    def title(self):
        return self.link.title


class Separator(Link):
    """
    Menu separator. Renders to an <hr> tag.
    """
    def __init__(self, *args, **kwargs):
        self._icon = None
        self.text = None
        self.view = None

    def resolve(self, *args, **kwargs):
        result = ResolvedLink(current_view_name=None, link=self)
        result.separator = True
        return result


class Text(Link):
    """
    Menu text. Renders to a plain <li> tag.
    """
    def __init__(self, *args, **kwargs):
        self.html_extra_classes = kwargs.get('html_extra_classes', '')
        self._icon = None
        self.text = kwargs.get('text')
        self.view = None

    def resolve(self, *args, **kwargs):
        result = ResolvedLink(current_view_name=None, link=self)
        result.context = kwargs.get('context')
        result.text_span = True
        return result



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig


class PlatformsForgeApp(MayanAppConfig):
    app_namespace = 'platforms_forge'
    app_url = 'platforms_forge'
    name = 'mayan.apps.platforms_forge'
    verbose_name = _(message='Platforms Forge')



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('document_parsing', '0009_auto_20201012_0708')
    ]

    operations = [
        migrations.AlterModelOptions(
            name='documentfilepagecontent',
            options={
                'verbose_name': 'Document file page content',
                'verbose_name_plural': 'Document file page contents'
            }
        ),
        migrations.AlterField(
            model_name='documentfilepagecontent',
            name='document_file_page',
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='content', to='documents.DocumentFilePage',
                verbose_name='Document file page'
            )
        )
    ]



from itertools import islice
import logging

from django.apps import apps
from django.db import models
from django.db.models import Q, Value

from mayan.apps.acls.models import AccessControlList
from mayan.apps.lock_manager.backends.base import LockingBackend
from mayan.apps.lock_manager.exceptions import LockError

from .classes import DuplicateBackend
from .literals import BULK_CREATE_BATCH_SIZE

logger = logging.getLogger(name=__name__)


class StoredDuplicateBackendManager(models.Manager):
    def scan_document(self, document):
        """
        Find duplicates of document based on each registered backend's logic.
        """
        lock_name = 'duplicates__scan_document-{}'.format(document.pk)
        try:
            logger.debug('trying to acquire lock: %s', lock_name)
            lock = LockingBackend.get_backend().acquire_lock(name=lock_name)
            logger.debug('acquired lock: %s', lock_name)
            try:
                DuplicateBackendEntry = apps.get_model(
                    app_label='duplicates', model_name='DuplicateBackendEntry'
                )

                for backend_path, backend_class in DuplicateBackend.get_all():
                    if backend_class.verify(document=document):

                        stored_backend, created = self.get_or_create(
                            backend_path=backend_path
                        )
                        duplicates = stored_backend.get_backend_instance().process(
                            document=document
                        )

                        if duplicates.exists():
                            duplicates_entry, created = stored_backend.duplicate_entries.get_or_create(
                                document=document
                            )

                            bulk_create_list = (
                                duplicates_entry.documents.through(
                                    duplicatebackendentry_id=duplicates_entry.pk,
                                    document=duplicate
                                ) for duplicate in duplicates
                            )

                            while True:
                                batch = list(
                                    islice(
                                        bulk_create_list,
                                        BULK_CREATE_BATCH_SIZE
                                    )
                                )

                                if not batch:
                                    break

                                duplicates_entry.documents.through.objects.bulk_create(
                                    batch_size=BULK_CREATE_BATCH_SIZE,
                                    ignore_conflicts=True, objs=batch
                                )

                            # Create empty duplicate entries for the
                            # duplicates as source that do not have one.
                            bulk_create_list = (
                                DuplicateBackendEntry(
                                    stored_backend=stored_backend,
                                    document=duplicate,
                                ) for duplicate in duplicates.filter(
                                    duplicates__stored_backend=None
                                )
                            )

                            while True:
                                batch = list(
                                    islice(
                                        bulk_create_list,
                                        BULK_CREATE_BATCH_SIZE
                                    )
                                )

                                if not batch:
                                    break

                                DuplicateBackendEntry.objects.bulk_create(
                                    batch_size=BULK_CREATE_BATCH_SIZE,
                                    ignore_conflicts=True, objs=batch
                                )

                            # Get all duplicate entries for the duplicates as
                            # source.
                            duplicate_entries = stored_backend.duplicate_entries.filter(
                                document__in=duplicates
                            )

                            # Create the many to many entries for this
                            # document as a target.
                            bulk_create_list = (
                                document.as_duplicate.through(
                                    duplicatebackendentry_id=duplicate_entry.pk,
                                    document_id=document.pk
                                ) for duplicate_entry in duplicate_entries
                            )

                            while True:
                                batch = list(
                                    islice(
                                        bulk_create_list,
                                        BULK_CREATE_BATCH_SIZE
                                    )
                                )

                                if not batch:
                                    break

                                document.as_duplicate.through.objects.bulk_create(
                                    batch_size=BULK_CREATE_BATCH_SIZE,
                                    ignore_conflicts=True, objs=batch
                                )
                        else:
                            # Document has no duplicates for this backend.
                            # Delete any existing entry for where this
                            # document is the source for the backend.
                            stored_backend.duplicate_entries.filter(
                                document=document
                            ).delete()

                            # Delete any existing entry for where this
                            # document is a duplicate for the backend.
                            document.as_duplicate.filter(
                                stored_backend=stored_backend
                            ).delete()
                    else:
                        # Document cannot be scanned for duplicates for this
                        # backend. Delete any existing entries for the
                        # backend where the document is a source or target.
                        stored_backend.duplicate_entries.filter(
                            documents=document
                        ).delete()

                        document.as_duplicate.filter(
                            stored_backend=stored_backend
                        ).delete()
            finally:
                lock.release()
        except LockError:
            logger.debug('unable to obtain lock: %s' % lock_name)
            raise


class DuplicateBackendEntryManager(models.Manager):
    def clean_empty_duplicate_lists(self):
        self.filter(documents=None).delete()

    def get_duplicated_documents(self, permission=None, user=None):
        DuplicateSourceDocument = apps.get_model(
            app_label='duplicates', model_name='DuplicateSourceDocument'
        )
        DuplicateTargetDocument = apps.get_model(
            app_label='duplicates', model_name='DuplicateTargetDocument'
        )

        queryset_target_documents = DuplicateTargetDocument.valid.all()

        if permission:
            queryset_target_documents = AccessControlList.objects.restrict_queryset(
                queryset=queryset_target_documents, permission=permission,
                user=user
            )

        queryset = self.filter(documents__in=queryset_target_documents).only(
            'document_id'
        ).values('document_id')

        return DuplicateSourceDocument.valid.filter(pk__in=queryset)

    def get_duplicates_of(self, document, permission=None, user=None):
        DuplicateTargetDocument = apps.get_model(
            app_label='duplicates', model_name='DuplicateTargetDocument'
        )

        queryset = DuplicateTargetDocument.valid.none()

        when_list = []

        for entry in self.filter(document=document).all():
            queryset = queryset | DuplicateTargetDocument.valid.filter(
                pk__in=entry.documents.only('pk').values('pk')
            )

            when_list.append(
                models.When(
                    Q(
                        pk__in=entry.documents.only('pk').values('pk')
                    ),
                    then=Value(
                        value=str(entry.stored_backend)
                    )
                )
            )

        queryset = queryset.annotate(
            backend=models.Case(
                *when_list, output_field=models.CharField()
            )
        )

        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset, user=user
            )

        return queryset



from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.class_mixins import MixinConditionTemplate
from mayan.apps.document_states.classes import WorkflowAction
from mayan.apps.documents.models.document_version_page_models import (
    DocumentVersionPage
)
from mayan.apps.templating.template_backends import Template

from .models import DocumentVersionPageOCRContent

__all__ = ('UpdateDocumentPageOCRAction',)


class UpdateDocumentPageOCRAction(MixinConditionTemplate, WorkflowAction):
    form_fields = {
        'page_condition': {
            'label': _(message='Page condition'),
            'class': 'mayan.apps.templating.fields.ModelTemplateField',
            'kwargs': {
                'initial_help_text': _(
                    message='The condition that will determine if a document '
                    'page\'s OCR content will be modified or not. The '
                    'condition is evaluated against the iterated document '
                    'page. Conditions that do not return any value, '
                    'that return the Python logical None, or an empty '
                    'string (\'\') are considered to be logical false, '
                    'any other value is considered to be the logical true.'
                ), 'required': False, 'model': DocumentVersionPage,
                'model_variable': 'document_page'
            }
        },
        'page_content': {
            'label': _(message='Page content'),
            'class': 'mayan.apps.templating.fields.ModelTemplateField',
            'kwargs': {
                'initial_help_text': _(
                    message='A template that will generate the OCR content '
                    'to be saved.'
                ), 'required': False, 'model': DocumentVersionPage,
                'model_variable': 'document_page'
            }
        }
    }
    label = _(message='Update document page OCR content')

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='OCR'), {
                    'fields': ('page_condition', 'page_content')
                },
            ),
        )
        return fieldsets

    def execute(self, context):
        document = context['workflow_instance'].document
        template = Template(
            template_string=self.kwargs['page_content']
        )

        for document_version_page in document.pages:
            context['document_version_page'] = document_version_page

            condition_result = self.evaluate_condition(context=context)

            if condition_result:
                template_result = template.render(context=context)

                DocumentVersionPageOCRContent.objects.update_or_create(
                    document_version_page=document_version_page, defaults={
                        'content': template_result
                    }
                )

    def get_condition_template(self):
        return self.kwargs['page_condition']



import logging
import os

import sh

from django.utils.translation import gettext_lazy as _

from ..classes import OCRBackendBase
from ..exceptions import OCRError

from .literals import DEFAULT_TESSERACT_BINARY_PATH, DEFAULT_TESSERACT_TIMEOUT

logger = logging.getLogger(name=__name__)


class Tesseract(OCRBackendBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.read_settings()

        if kwargs.get('auto_initialize', True):
            self.initialize()

    def _execute(self, image_file_object):
        """
        Execute the command line binary of tesseract.
        """
        if self.command_tesseract:
            arguments = ['-', '-']

            keyword_arguments = {
                '_in': image_file_object,
                '_timeout': self.command_timeout
            }

            if self.language:
                keyword_arguments['l'] = self.language

            environment = os.environ.copy()
            environment.update(self.command_environment)
            keyword_arguments['_env'] = environment

            arguments.extend(self.tesseract_arguments_extra)

            logger.debug(
                'Calling Tesseract with arguments %s, %s', arguments,
                keyword_arguments
            )

            try:
                output = self.command_tesseract(
                    *arguments, **keyword_arguments
                )
            except Exception as exception:
                error_message_list = []
                error_message_list.append(
                    'Exception calling Tesseract with language option: {}; {}'.format(
                        self.language, exception
                    )
                )

                if self.language not in self.languages:
                    error_message_list.append(
                        'The requested OCR language "{}" is not '
                        'available and needs to be installed.'.format(
                            self.language
                        )
                    )

                error_message = '/n'.join(error_message_list)

                logger.error(error_message, exc_info=True)
                raise OCRError(error_message)
            else:
                return output
        else:
            return ''

    def initialize(self):
        self.languages = ()

        try:
            self.command_tesseract = sh.Command(
                path=self.tesseract_binary_path
            )
        except sh.CommandNotFound:
            self.command_tesseract = None
            raise OCRError(
                _(message='Tesseract OCR not found.')
            )
        else:
            # Get version.
            output = self.command_tesseract(v=True)
            logger.debug('Tesseract version: %s', output)

            # Get languages.
            output = self.command_tesseract(list_langs=True)
            # Sample output format.
            # List of available languages (3):
            # deu
            # eng
            # osd
            # <- empty line

            # Extraction: strip last line, split by newline, discard the
            # first line.
            self.languages = output.strip().split('\n')[1:]

            logger.debug(
                'Available languages: %s', ', '.join(self.languages)
            )

    def read_settings(self):
        self.command_timeout = self.kwargs.get(
            'timeout', DEFAULT_TESSERACT_TIMEOUT
        )
        self.command_environment = self.kwargs.get(
            'environment', {}
        )
        self.tesseract_arguments_extra = self.kwargs.get(
            'tesseract_arguments_extra', {}
        )
        self.tesseract_binary_path = self.kwargs.get(
            'tesseract_path', DEFAULT_TESSERACT_BINARY_PATH
        )



import hashlib
import io
import logging

from PIL import Image
from furl import furl

from django.apps import apps
from django.db.models import Max
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.serialization import yaml_load
from mayan.apps.file_caching.models import CachePartitionFile

from .literals import STORAGE_NAME_ASSETS_CACHE
from .transformations import BaseTransformation

logger = logging.getLogger(name=__name__)


class AssetBusinessLogicMixin:
    @cached_property
    def cache(self):
        Cache = apps.get_model(app_label='file_caching', model_name='Cache')

        return Cache.objects.get(
            defined_storage_name=STORAGE_NAME_ASSETS_CACHE
        )

    @cached_property
    def cache_partition(self):
        partition, created = self.cache.partitions.get_or_create(
            name='{}'.format(self.pk)
        )
        return partition

    def generate_image(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None
    ):
        # The parameters 'maximum_layer_order',
        # `transformation_instance_list`, `user` are not used, but added
        # to retain interface compatibility.
        cache_filename = '{}'.format(
            self.get_hash()
        )

        try:
            self.cache_partition.get_file(filename=cache_filename)
        except CachePartitionFile.DoesNotExist:
            logger.debug('asset cache file "%s" not found', cache_filename)

            image = self.get_image()
            with io.BytesIO() as image_buffer:
                image.save(image_buffer, format='PNG')

                with self.cache_partition.create_file(filename=cache_filename) as file_object:
                    file_object.write(
                        image_buffer.getvalue()
                    )
        else:
            logger.debug('asset cache file "%s" found', cache_filename)

        return cache_filename

    def get_api_image_url(self, *args, **kwargs):
        final_url = furl()
        final_url.args = kwargs
        final_url.path = reverse(
            viewname='rest_api:asset-image',
            kwargs={'asset_id': self.pk}
        )
        final_url.args['_hash'] = self.get_hash()

        return final_url.tostr()

    def get_hash(self):
        with self.open() as file_object:
            hash_object = hashlib.sha256(
                string=file_object.read()
            )
            return hash_object.hexdigest()

    def get_image(self):
        with self.open() as file_object:
            image = Image.open(fp=file_object)
            image.load()

            if image.mode != 'RGBA':
                image.putalpha(alpha=255)

        return image

    def open(self):
        name = self.file.name
        self.file.close()
        return self.file.storage.open(name=name)


class ObjectLayerBusinessLogicMixin:
    def get_next_order(self):
        last_order = self.transformations.aggregate(
            Max('order')
        )['order__max']

        if last_order is not None:
            return last_order + 1
        else:
            return 0


class LayerTransformationBusinessLogicMixin:
    def get_arguments_column(self):
        try:
            arguments = yaml_load(stream=self.arguments or '{}')
        except Exception:
            arguments = {
                'error': _(message='Badly formatted arguments YAML')
            }

        result = []
        for key, value in arguments.items():
            result.append(
                '{}: {}'.format(key, value)
            )

        return ', '.join(result)

    get_arguments_column.short_description = _(message='Arguments')

    def get_transformation_class(self):
        return BaseTransformation.get(name=self.name)



from django.utils.translation import gettext_lazy as _

from mayan.apps.forms import form_fields, forms

from ..models import WorkflowTransition


class WorkflowInstanceTransitionSelectForm(forms.Form):
    def __init__(self, user, workflow_instance, *args, **kwargs):
        super().__init__(*args, **kwargs)

        queryset = workflow_instance.get_queryset_valid_transitions(user=user)
        self.fields['transition'].queryset = queryset

    transition = form_fields.ModelChoiceField(
        help_text=_(
            message='Select a transition to execute in the next step.'
        ), label=_(message='Transition'),
        queryset=WorkflowTransition.objects.none()
    )



from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import FilteredPrimaryKeyRelatedField

from ..models import Workflow


class WorkflowTemplateSerializer(serializers.HyperlinkedModelSerializer):
    document_types_add_url = serializers.HyperlinkedIdentityField(
        label=_(message='Document types add URL'),
        lookup_url_kwarg='workflow_template_id',
        view_name='rest_api:workflow-template-document-type-add'
    )
    document_types_remove_url = serializers.HyperlinkedIdentityField(
        label=_(message='Document types remove URL'),
        lookup_url_kwarg='workflow_template_id',
        view_name='rest_api:workflow-template-document-type-remove'
    )
    document_types_url = serializers.HyperlinkedIdentityField(
        label=_(message='Document types URL'),
        lookup_url_kwarg='workflow_template_id',
        view_name='rest_api:workflow-template-document-type-list'
    )
    documents_url = serializers.HyperlinkedIdentityField(
        label=_(message='Documents URL'),
        lookup_url_kwarg='workflow_template_id',
        view_name='rest_api:workflow-template-document-list'
    )
    image_url = serializers.HyperlinkedIdentityField(
        label=_(message='Image URL'), lookup_url_kwarg='workflow_template_id',
        view_name='rest_api:workflow-template-image'
    )
    states_url = serializers.HyperlinkedIdentityField(
        label=_(message='States URL'), lookup_url_kwarg='workflow_template_id',
        view_name='rest_api:workflow-template-state-list'
    )
    transitions_url = serializers.HyperlinkedIdentityField(
        label=_(message='Transitions URL'), lookup_url_kwarg='workflow_template_id',
        view_name='rest_api:workflow-template-transition-list'
    )

    class Meta:
        extra_kwargs = {
            'url': {
                'label': _(message='URL'),
                'lookup_url_kwarg': 'workflow_template_id',
                'view_name': 'rest_api:workflow-template-detail'
            }
        }
        fields = (
            'auto_launch', 'document_types_add_url',
            'document_types_remove_url', 'document_types_url',
            'documents_url', 'id', 'ignore_completed', 'image_url',
            'internal_name', 'label', 'states_url', 'transitions_url', 'url'
        )
        model = Workflow
        read_only_fields = (
            'document_types_add_url', 'document_types_remove_url',
            'document_types_url', 'documents_url', 'id', 'image_url',
            'states_url', 'transitions_url', 'url'
        )


class WorkflowTemplateDocumentTypeAddSerializer(serializers.Serializer):
    document_type_id = FilteredPrimaryKeyRelatedField(
        help_text=_(
            message='Primary key of the document type to add to the workflow.'
        ), label=_(message='Document type ID'), source_model=DocumentType,
        source_permission=permission_document_type_edit
    )


class WorkflowTemplateDocumentTypeRemoveSerializer(serializers.Serializer):
    document_type_id = FilteredPrimaryKeyRelatedField(
        help_text=_(
            message='Primary key of the document type to remove from the workflow.'
        ), label=_(message='Document type ID'), source_model=DocumentType,
        source_permission=permission_document_type_edit
    )



from django.utils.translation import gettext_lazy as _

from mayan.apps.forms import form_fields, form_widgets, forms

from ..literals import PAGE_RANGE_ALL, PAGE_RANGE_CHOICES


class PrintForm(forms.Form):
    page_group = form_fields.ChoiceField(
        choices=PAGE_RANGE_CHOICES, initial=PAGE_RANGE_ALL,
        label=_(message='Page group'), widget=form_widgets.RadioSelect
    )
    page_range = form_fields.CharField(
        label=_(message='Page range'), required=False
    )


class PageNumberForm(forms.Form):
    page = form_fields.ModelChoiceField(
        help_text=_(
            message='Page number from which all the transformations will be cloned. '
            'Existing transformations will be lost.'
        ), queryset=None
    )

    def __init__(self, *args, **kwargs):
        self.instance = kwargs.pop('instance')
        super().__init__(*args, **kwargs)
        self.fields['page'].queryset = self.instance.pages.all()



from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.template import RequestContext
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.forms import forms
from mayan.apps.views.generics import (
    FormView, MultipleObjectDeleteView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from ..forms.workflow_instance_forms import (
    WorkflowInstanceTransitionSelectForm
)
from ..icons import (
    icon_workflow_instance_delete, icon_workflow_instance_detail,
    icon_workflow_instance_list, icon_workflow_instance_transition,
    icon_workflow_instance_transition_select, icon_workflow_template_list
)
from ..links import link_workflow_instance_transition
from ..models import WorkflowInstance
from ..permissions import (
    permission_workflow_instance_delete,
    permission_workflow_instance_transition, permission_workflow_template_view
)


class WorkflowInstanceListView(ExternalObjectViewMixin, SingleObjectListView):
    external_object_permission = permission_workflow_template_view
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    object_permission = permission_workflow_template_view
    view_icon = icon_workflow_instance_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'no_results_icon': icon_workflow_template_list,
            'no_results_text': _(
                message='Assign workflows to the document type of this document '
                'to have this document execute those workflows. '
            ),
            'no_results_title': _(
                message='There are no workflows for this document'
            ),
            'object': self.external_object,
            'title': _(
                message='Workflows for document: %s'
            ) % self.external_object
        }

    def get_source_queryset(self):
        return self.external_object.workflows.all()


class WorkflowInstanceDeleteView(MultipleObjectDeleteView):
    object_permission = permission_workflow_instance_delete
    pk_url_kwarg = 'workflow_instance_id'
    success_message_plural = _(
        message='%(count)d workflow instances deleted successfully.'
    )
    success_message_single = _(
        message='Workflow instance "%(object)s" deleted successfully.'
    )
    success_message_singular = _(
        message='%(count)d workflow instance deleted successfully.'
    )
    title_plural = _(
        message='Delete the %(count)d selected workflow instances'
    )
    title_single = _(message='Delete workflow instance: %(object)s')
    title_singular = _(
        message='Delete the %(count)d selected workflow instance'
    )
    view_icon = icon_workflow_instance_delete

    def get_extra_context(self):
        return {
            'navigation_object_list': ('document', 'workflow_instance'),
            'document': self.object.document,
            'workflow_instance': self.object
        }

    def get_post_action_redirect(self):
        # Use [0] instead of first(). First returns None and it is not
        # usable.
        return reverse(
            kwargs={
                'document_id': self.object_list[0].document_id
            }, viewname='document_states:workflow_instance_list'
        )

    def get_source_queryset(self):
        queryset_documents = AccessControlList.objects.restrict_queryset(
            queryset=Document.valid.all(),
            permission=permission_workflow_instance_delete,
            user=self.request.user
        )

        queryset_workflow_instances = WorkflowInstance.objects.filter(
            document_id__in=queryset_documents.values('pk')
        )

        return queryset_workflow_instances


class WorkflowInstanceDetailView(
    ExternalObjectViewMixin, SingleObjectListView
):
    external_object_permission = permission_workflow_template_view
    external_object_pk_url_kwarg = 'workflow_instance_id'
    object_permission = permission_workflow_template_view
    view_icon = icon_workflow_instance_detail

    def get_extra_context(self):
        return {
            'hide_object': True,
            'navigation_object_list': ('object', 'workflow_instance'),
            'no_results_icon': icon_workflow_instance_detail,
            'no_results_main_link': link_workflow_instance_transition.resolve(
                context=RequestContext(
                    dict_={'object': self.external_object},
                    request=self.request
                )
            ),
            'no_results_text': _(
                message='This view will show the state changes as a workflow '
                'instance is transitioned.'
            ),
            'no_results_title': _(
                message='There are no details for this workflow instance'
            ),
            'object': self.external_object.document,
            'title': _(message='Detail of workflow: %(workflow)s') % {
                'workflow': self.external_object
            },
            'workflow_instance': self.external_object
        }

    def get_external_object_queryset(self):
        queryset_documents = AccessControlList.objects.restrict_queryset(
            queryset=Document.valid.all(),
            permission=permission_workflow_template_view,
            user=self.request.user
        )

        return WorkflowInstance.objects.filter(
            document_id__in=queryset_documents.values('pk')
        )

    def get_source_queryset(self):
        return self.external_object.log_entries.order_by('-datetime')


class WorkflowInstanceTransitionExecuteView(
    ExternalObjectViewMixin, FormView
):
    external_object_pk_url_kwarg = 'workflow_instance_id'
    external_object_queryset = WorkflowInstance.valid.all()
    form_class = forms.DynamicForm
    template_name = 'appearance/form_container.html'
    view_icon = icon_workflow_instance_transition

    def form_valid(self, form):
        form_data = form.cleaned_data
        comment = form_data.pop('comment')

        self.external_object.do_transition(
            comment=comment, extra_data=form_data,
            transition=self.get_workflow_template_transition(),
            user=self.request.user
        )
        messages.success(
            message=_(
                message='Document "%s" transitioned successfully'
            ) % self.external_object.document, request=self.request
        )
        return HttpResponseRedirect(
            redirect_to=self.get_success_url()
        )

    def get_external_object_queryset_filtered(self):
        queryset = super().get_external_object_queryset_filtered()

        # Filter further down by document access.
        queryset_documents = AccessControlList.objects.restrict_queryset(
            permission=permission_workflow_instance_transition,
            queryset=Document.valid.all(),
            user=self.request.user
        )

        return queryset.filter(document__in=queryset_documents)

    def get_extra_context(self):
        return {
            'navigation_object_list': ('object', 'workflow_instance'),
            'object': self.external_object.document,
            'title': _(
                message='Execute transition "%(transition)s" for workflow: '
                '%(workflow)s'
            ) % {
                'transition': self.get_workflow_template_transition(),
                'workflow': self.external_object
            },
            'workflow_instance': self.external_object
        }

    def get_form_extra_kwargs(self):
        schema = {
            'fields': {
                'comment': {
                    'label': _(message='Comment'),
                    'class': 'django.forms.CharField', 'kwargs': {
                        'help_text': _(
                            message='Optional comment to attach to the '
                            'transition.'
                        ),
                        'required': False,
                    }
                }
            },
            'widgets': {
                'comment': {
                    'class': 'django.forms.widgets.Textarea',
                    'kwargs': {
                        'attrs': {'rows': 3}
                    }
                }
            }
        }

        workflow_template_transition = self.get_workflow_template_transition()
        workflow_template_transition.get_form_schema(
            schema=schema, workflow_instance=self.external_object
        )

        return {'schema': schema}

    def get_success_url(self):
        return self.external_object.get_absolute_url()

    def get_workflow_template_transition(self):
        return get_object_or_404(
            klass=self.external_object.get_queryset_valid_transitions(
                user=self.request.user
            ), pk=self.kwargs['workflow_template_transition_id']
        )


class WorkflowInstanceTransitionSelectView(
    ExternalObjectViewMixin, FormView
):
    external_object_pk_url_kwarg = 'workflow_instance_id'
    external_object_queryset = WorkflowInstance.valid.all()
    form_class = WorkflowInstanceTransitionSelectForm
    template_name = 'appearance/form_container.html'
    view_icon = icon_workflow_instance_transition_select

    def form_valid(self, form):
        return HttpResponseRedirect(
            redirect_to=reverse(
                kwargs={
                    'workflow_instance_id': self.external_object.pk,
                    'workflow_template_transition_id': form.cleaned_data['transition'].pk
                },
                viewname='document_states:workflow_instance_transition_execute'
            )
        )

    def get_extra_context(self):
        return {
            'navigation_object_list': ('object', 'workflow_instance'),
            'object': self.external_object.document,
            'submit_label': _(message='Select'),
            'title': _(
                message='Select transition for workflow "%(workflow)s" of document "%(document)s"'
            ) % {
                'document': self.external_object.document,
                'workflow': self.external_object
            },
            'workflow_instance': self.external_object
        }

    def get_external_object_queryset_filtered(self):
        queryset = super().get_external_object_queryset_filtered()

        # Filter further down by document access.
        queryset_documents = AccessControlList.objects.restrict_queryset(
            permission=permission_workflow_instance_transition,
            queryset=Document.valid.all(),
            user=self.request.user
        )

        return queryset.filter(document__in=queryset_documents)

    def get_form_extra_kwargs(self):
        return {
            'user': self.request.user,
            'workflow_instance': self.external_object
        }



from django.utils.translation import gettext_lazy as _

from mayan.apps.forms import forms
from mayan.apps.templating.fields import ModelTemplateField

from ..models.workflow_instance_models import WorkflowInstance
from ..models.workflow_transition_field_models import WorkflowTransitionField


class WorkflowTransitionFieldForm(forms.ModelForm):
    fieldsets = (
        (
            _(message='Basic'), {
                'fields': ('name', 'label')
            }
        ), (
            _(message='Field'), {
                'fields': ('field_type', 'help_text')
            }
        ), (
            _(message='Value'), {
                'fields': ('default', 'lookup', 'required')
            }
        ), (
            _(message='Appearance'), {
                'fields': ('widget', 'widget_kwargs')
            }
        )
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields['default'] = ModelTemplateField(
            initial_help_text=self.fields['default'].help_text,
            model=WorkflowInstance, model_variable='workflow_instance',
            required=False
        )
        self.fields['lookup'] = ModelTemplateField(
            context_entry_name_list=('groups', 'users'),
            initial_help_text=self.fields['lookup'].help_text,
            model=WorkflowInstance, model_variable='workflow_instance',
            required=False
        )

    class Meta:
        fields = (
            'default', 'field_type', 'help_text', 'label', 'lookup', 'name',
            'required', 'widget', 'widget_kwargs'
        )
        model = WorkflowTransitionField



from mayan.apps.rest_api import generics

from ..models.metadata_type_models import MetadataType
from ..permissions import (
    permission_metadata_type_create, permission_metadata_type_delete,
    permission_metadata_type_edit, permission_metadata_type_view
)
from ..serializers import MetadataTypeSerializer


class APIMetadataTypeListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the metadata types.
    post: Create a new metadata type.
    """
    mayan_object_permission_map = {'GET': permission_metadata_type_view}
    mayan_view_permission_map = {'POST': permission_metadata_type_create}
    serializer_class = MetadataTypeSerializer
    source_queryset = MetadataType.objects.all()

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class APIMetadataTypeView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected metadata type.
    get: Return the details of the selected metadata type.
    patch: Edit the selected metadata type.
    put: Edit the selected metadata type.
    """
    lookup_url_kwarg = 'metadata_type_id'
    mayan_object_permission_map = {
        'DELETE': permission_metadata_type_delete,
        'GET': permission_metadata_type_view,
        'PATCH': permission_metadata_type_edit,
        'PUT': permission_metadata_type_edit
    }
    serializer_class = MetadataTypeSerializer
    source_queryset = MetadataType.objects.all()

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('mailer', '0002_usermailer_usermailerlogentry')
    ]

    operations = [
        migrations.AddField(
            model_name='usermailer',
            name='enabled',
            field=models.BooleanField(default=True, verbose_name='Enabled'),
        ),
        migrations.AlterField(
            model_name='usermailerlogentry',
            name='user_mailer',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='error_log', to='mailer.UserMailer',
                verbose_name='User mailer'
            )
        )
    ]



import whoosh
from whoosh.qparser import FuzzyTermPlugin, GtLtPlugin, RegexPlugin
from whoosh.qparser.dateparse import DateParserPlugin

from ...search_query_types import (
    BackendQueryType, QueryTypeExact, QueryTypeFuzzy, QueryTypeGreaterThan,
    QueryTypeGreaterThanOrEqual, QueryTypeLessThan, QueryTypeLessThanOrEqual,
    QueryTypePartial, QueryTypeRange, QueryTypeRangeExclusive,
    QueryTypeRegularExpression
)

from .backend import WhooshSearchBackend


class BackendQueryTypeWhoosh(BackendQueryType):
    def do_resolve(self):
        if self.get_search_backend_field_type() == whoosh.fields.DATETIME:

            self.extra_kwargs['parser'].add_plugin(
                DateParserPlugin()
            )

        return self._do_resolve()


class BackendQueryTypeExact(BackendQueryTypeWhoosh):
    query_type = QueryTypeExact

    def _do_resolve(self):
        if self.value is not None:
            if not self.value and self.is_quoted_value:
                return 'NOT *'
            else:
                if self.is_quoted_value:
                    template = '{}:("{}")'
                else:
                    template = '{}:({})'

                return template.format(
                    self.search_field.field_name, self.value
                )


class BackendQueryTypeFuzzy(BackendQueryTypeWhoosh):
    query_type = QueryTypeFuzzy

    def _do_resolve(self):
        self.extra_kwargs['parser'].add_plugin(
            FuzzyTermPlugin()
        )

        if self.value is not None:
            if self.is_quoted_value:
                template = '{}:"{}"~2'
            else:
                template = '{}:{}~2'

            return template.format(
                self.search_field.field_name, self.value
            )


class BackendQueryTypeComparison(BackendQueryTypeWhoosh):
    def _do_resolve(self):
        self.extra_kwargs['parser'].add_plugin(
            GtLtPlugin()
        )

        if self.value is not None:
            return self.template.format(
                self.search_field.field_name, self.value
            )


class BackendQueryTypeGreaterThan(BackendQueryTypeComparison):
    query_type = QueryTypeGreaterThan
    template = '{}:>{}'


class BackendQueryTypeGreaterThanOrEqual(BackendQueryTypeComparison):
    query_type = QueryTypeGreaterThanOrEqual
    template = '{}:>={}'


class BackendQueryTypeLessThan(BackendQueryTypeComparison):
    query_type = QueryTypeLessThan
    template = '{}:<{}'


class BackendQueryTypeLessThanOrEqual(BackendQueryTypeComparison):
    query_type = QueryTypeLessThanOrEqual
    template = '{}:<={}'


class BackendQueryTypePartial(BackendQueryTypeWhoosh):
    query_type = QueryTypePartial

    def _do_resolve(self):
        if self.value is not None:
            if self.is_quoted_value:
                template = '{}:"{}"'
            else:
                template = '{}:(*{}*)'

            if self.get_search_backend_field_type() == whoosh.fields.BOOLEAN:
                template = '{}:({})'

            if self.value is not None:
                return template.format(
                    self.search_field.field_name, self.value
                )


class BackendQueryTypeRange(BackendQueryTypeWhoosh):
    query_type = QueryTypeRange

    def _do_resolve(self):
        if self.value is not None:
            return '{}:[{} TO {}]'.format(
                self.search_field.field_name, *self.value
            )


class BackendQueryTypeRangeExclusive(BackendQueryTypeWhoosh):
    query_type = QueryTypeRangeExclusive

    def _do_resolve(self):
        if self.value is not None:
            return '{}:{{{} TO {}}}'.format(
                self.search_field.field_name, *self.value
            )


class BackendQueryTypeRegularExpression(BackendQueryTypeWhoosh):
    query_type = QueryTypeRegularExpression

    def _do_resolve(self):
        self.extra_kwargs['parser'].add_plugin(
            RegexPlugin()
        )

        if self.value is not None:
            return '{}:r"{}"'.format(
                self.search_field.field_name, self.value
            )


BackendQueryType.register(
    klass=BackendQueryTypeExact, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeFuzzy, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypePartial, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeGreaterThan, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeGreaterThanOrEqual,
    search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeLessThan, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeLessThanOrEqual, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRange, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRangeExclusive, search_backend=WhooshSearchBackend
)
BackendQueryType.register(
    klass=BackendQueryTypeRegularExpression,
    search_backend=WhooshSearchBackend
)



from django.utils.functional import cached_property

from .classes import EventType
from .literals import TEXT_UNKNOWN_EVENT_ID


class StoredEventTypeBusinessLogicMixin:
    @cached_property
    def event_type(self):
        return EventType.get(id=self.name)

    @property
    def label(self):
        try:
            event_type = self.event_type
        except KeyError:
            return TEXT_UNKNOWN_EVENT_ID % self.name
        else:
            return event_type.label

    @property
    def namespace(self):
        return self.event_type.namespace


class NotificationBusinessLogicMixin:
    def get_event_type(self):
        try:
            return EventType.get(id=self.action.verb)
        except KeyError:
            return None



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig
from mayan.apps.templating.classes import AJAXTemplate

from .handlers import handler_pre_initial_setup, handler_pre_upgrade
from .links import (
    link_about, link_knowledge_base, link_license, link_separator_information,
    link_setup, link_support, link_tools
)
from .menus import menu_system, menu_topbar, menu_user
from .settings import setting_home_view
from .signals import signal_pre_initial_setup, signal_pre_upgrade


class CommonApp(MayanAppConfig):
    app_namespace = 'common'
    app_url = ''
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.common'
    static_media_ignore_patterns = (
        'mptt/*',
    )
    verbose_name = _(message='Common')

    def ready(self):
        super().ready()

        AJAXTemplate(
            name='menu_main', template_name='appearance/menus/main.html'
        )
        AJAXTemplate(
            context={'home_view': setting_home_view.value},
            name='menu_topbar',
            template_name='appearance/menus/topbar.html'
        )

        menu_system.bind_links(
            links=(
                link_tools, link_setup, link_separator_information,
                link_knowledge_base, link_support, link_about, link_license
            )
        )
        menu_topbar.bind_links(
            links=(menu_system, menu_user),
            position=10
        )

        signal_pre_initial_setup.connect(
            dispatch_uid='common_handler_pre_initial_setup',
            receiver=handler_pre_initial_setup
        )
        signal_pre_upgrade.connect(
            dispatch_uid='common_handler_pre_upgrade',
            receiver=handler_pre_upgrade
        )



from django.db import migrations, models

import mayan.apps.converter.models


class Migration(migrations.Migration):
    dependencies = [
        ('converter', '0005_auto_20150708_0118')
    ]

    operations = [
        migrations.AlterField(
            model_name='transformation',
            name='arguments',
            field=models.TextField(
                help_text='Enter the arguments for the transformation as a '
                'YAML dictionary. ie: {"degrees": 180}', blank=True,
                verbose_name='Arguments', validators=getattr(
                    mayan.apps.converter.models, 'validators', []
                )
            ),
            preserve_default=True
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig
from mayan.apps.common.signals import signal_post_initial_setup


class SourceWebFormsApp(MayanAppConfig):
    app_namespace = 'source_web_forms'
    app_url = 'source_web_forms'
    has_tests = True
    name = 'mayan.apps.source_web_forms'
    verbose_name = _(message='Web form sources')

    def ready(self):
        super().ready()

        # Hidden import to avoid errors attempting to load models before
        # they are ready.
        from .handlers import handler_create_default_document_source

        signal_post_initial_setup.connect(
            receiver=handler_create_default_document_source,
            dispatch_uid='sources_web_forms_handler_create_default_document_source'
        )



from django.db import models
from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import (
    EventManagerMethodAfter, EventManagerSave
)
from mayan.apps.events.models import StoredEventType

from ..events import event_workflow_template_edited
from ..managers import WorkflowTransitionTriggerEventManager

from .workflow_transition_models import WorkflowTransition
from .workflow_transition_trigger_model_mixins import (
    WorkflowTransitionTriggerEventBusinessLogicMixin
)

__all__ = ('WorkflowTransitionTriggerEvent',)


class WorkflowTransitionTriggerEvent(
    ExtraDataModelMixin, WorkflowTransitionTriggerEventBusinessLogicMixin,
    models.Model
):
    transition = models.ForeignKey(
        on_delete=models.CASCADE, related_name='trigger_events',
        to=WorkflowTransition, verbose_name=_(message='Transition')
    )
    event_type = models.ForeignKey(
        on_delete=models.CASCADE, to=StoredEventType,
        verbose_name=_(message='Event type')
    )

    objects = WorkflowTransitionTriggerEventManager()

    class Meta:
        ordering = ('event_type__name',)
        unique_together = ('transition', 'event_type')
        verbose_name = _(message='Workflow transition trigger event')
        verbose_name_plural = _(
            message='Workflow transitions trigger events'
        )

    def __str__(self):
        return str(self.transition)

    @method_event(
        action_object='self',
        event_manager_class=EventManagerMethodAfter,
        event=event_workflow_template_edited,
        target='transition.workflow'
    )
    def delete(self, *args, **kwargs):
        return super().delete(*args, **kwargs)

    @method_event(
        event_manager_class=EventManagerSave,
        created={
            'action_object': 'self',
            'event': event_workflow_template_edited,
            'target': 'transition.workflow'
        },
        edited={
            'action_object': 'self',
            'event': event_workflow_template_edited,
            'target': 'transition.workflow'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)



from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.models.document_models import Document
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import (
    EventManagerMethodAfter, EventManagerSave
)

from ..events import (
    event_document_metadata_added, event_document_metadata_edited,
    event_document_metadata_removed
)

from .metadata_instance_model_mixins import DocumentMetadataBusinessLogicMixin
from .metadata_type_models import MetadataType


class DocumentMetadata(
    DocumentMetadataBusinessLogicMixin, ExtraDataModelMixin, models.Model
):
    """
    Model used to link an instance of a metadata type with a value to a
    document.
    """
    _ordering_fields = ('value',)

    document = models.ForeignKey(
        on_delete=models.CASCADE, related_name='metadata', to=Document,
        verbose_name=_(message='Document')
    )
    metadata_type = models.ForeignKey(
        on_delete=models.CASCADE, to=MetadataType, verbose_name=_(
            message='Type'
        )
    )
    value = models.TextField(
        blank=True, help_text=_(
            message='The actual value stored in the metadata type field for '
            'the document.'
        ), null=True, verbose_name=_(message='Value')
    )

    class Meta:
        ordering = ('metadata_type',)
        unique_together = ('document', 'metadata_type')
        verbose_name = _(message='Document metadata')
        verbose_name_plural = _(message='Document metadata')

    def __str__(self):
        return str(self.metadata_type)

    def clean_fields(self, *args, **kwargs):
        """
        Pass the value of the metadata being created to the parsers and
        validators for cleanup before saving.
        """
        super().clean_fields(*args, **kwargs)

        self.value = self.metadata_type.validate_value(
            document_type=self.document.document_type, value=self.value
        )

    @method_event(
        action_object='metadata_type',
        event=event_document_metadata_removed,
        event_manager_class=EventManagerMethodAfter,
        target='document'
    )
    def delete(self, enforce_required=True, *args, **kwargs):
        """
        Delete a metadata from a document. enforce_required which defaults
        to True, prevents deletion of required metadata at the model level.
        It used set to False when deleting document metadata on document
        type change.
        """
        is_required_for_document_type = enforce_required and self.document.document_type.metadata.filter(
            required=True
        ).filter(metadata_type=self.metadata_type).exists()

        if is_required_for_document_type:
            raise ValidationError(
                message=_(
                    message='Metadata type is required for this document '
                    'type.'
                )
            )

        return super().delete(*args, **kwargs)

    def natural_key(self):
        return self.document.natural_key() + self.metadata_type.natural_key()
    natural_key.dependencies = [
        'documents.Document', 'metadata.MetadataType'
    ]

    @method_event(
        created={
            'action_object': 'metadata_type',
            'event': event_document_metadata_added,
            'target': 'document'
        },
        edited={
            'action_object': 'metadata_type',
            'event': event_document_metadata_edited,
            'target': 'document'
        },
        event_manager_class=EventManagerSave
    )
    def save(self, *args, **kwargs):
        is_not_valid_for_document_type = not self.document.document_type.metadata.filter(
            metadata_type=self.metadata_type
        ).exists()

        if is_not_valid_for_document_type:
            raise ValidationError(
                message=_(
                    message='Metadata type is not valid for this document '
                    'type.'
                )
            )

        return super().save(*args, **kwargs)


class DocumentMetadataSearchResult(DocumentMetadata):
    class Meta:
        proxy = True



import logging

from django.utils.translation import gettext_lazy as _

from mayan.apps.django_gpg.models import Key
from mayan.apps.django_gpg.permissions import permission_key_sign
from mayan.apps.document_states.classes import WorkflowAction
from mayan.apps.document_states.exceptions import WorkflowStateActionError

from .models import DetachedSignature, EmbeddedSignature

logger = logging.getLogger(name=__name__)


class DocumentSignatureDetachedAction(WorkflowAction):
    form_field_widgets = {
        'key': {
            'class': 'django.forms.widgets.Select', 'kwargs': {
                'attrs': {'class': 'select2'}
            }
        },
        'passphrase': {
            'class': 'django.forms.widgets.PasswordInput',
        }
    }
    form_fields = {
        'passphrase': {
            'label': _(message='Passphrase'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    message='The passphrase to unlock the key and allow it to be '
                    'used to sign the document file.'
                ), 'required': False
            }
        }
    }
    label = _(message='Sign document (detached)')

    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'key': {
                    'class': 'mayan.apps.forms.form_fields.FormFieldFilteredModelChoice',
                    'help_text': _(
                        message='Private key that will be used to sign the document '
                        'file.'
                    ),
                    'kwargs': {
                        'source_model': Key,
                        'permission': permission_key_sign
                    },
                    'label': _(message='Private key'),
                    'required': True
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='Key'), {
                    'fields': ('key', 'passphrase',)
                }
            ),
        )
        return fieldsets

    def get_arguments(self, context):
        latest_file = context['workflow_instance'].document.file_latest
        if not latest_file:
            raise WorkflowStateActionError(
                _(
                    message='Document has no file to sign. You might be trying to '
                    'use this action in an initial state before the '
                    'created document is yet to be processed.'
                )
            )

        return {
            'document_file': latest_file,
            'key': Key.objects.get(
                pk=self.kwargs['key']
            ),
            'passphrase': self.kwargs.get('passphrase')
        }

    def execute(self, context):
        kwargs = self.get_arguments(context=context)

        DetachedSignature.objects.sign_document_file(**kwargs)


class DocumentSignatureEmbeddedAction(DocumentSignatureDetachedAction):
    label = _(message='Sign document (embedded)')

    def execute(self, context):
        kwargs = self.get_arguments(context=context)

        EmbeddedSignature.objects.sign_document_file(**kwargs)



from mayan.apps.rest_api import generics

from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.serializers.document_serializers import (
    DocumentSerializer
)

from ..models.workflow_state_models import WorkflowStateRuntimeProxy
from ..permissions import (
    permission_workflow_template_edit, permission_workflow_template_view
)
from ..serializers.workflow_template_state_serializers import (
    WorkflowTemplateStateActionSerializer, WorkflowTemplateStateSerializer
)

from .api_view_mixins import (
    ParentObjectWorkflowTemplateAPIViewMixin,
    ParentObjectWorkflowTemplateStateAPIViewMixin
)


class APIWorkflowTemplateStateDocumentListView(
    ParentObjectWorkflowTemplateStateAPIViewMixin, generics.ListAPIView
):
    """
    get: Return a list of all the documents at a specific workflow template state.
    """
    mayan_external_object_permission_map = {
        'GET': permission_workflow_template_view
    }
    mayan_object_permission_map = {'GET': permission_document_view}
    serializer_class = DocumentSerializer

    def get_source_queryset(self):
        workflow_template_state = self.get_workflow_template_state()

        workflow_template_state_runtime_proxy = WorkflowStateRuntimeProxy.objects.get(
            pk=workflow_template_state.pk
        )

        return workflow_template_state_runtime_proxy.get_documents().all()


class APIWorkflowTemplateStateListView(
    ParentObjectWorkflowTemplateAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of all the workflow template states.
    post: Create a new workflow template state.
    """
    mayan_external_object_permission_map = {
        'GET': permission_workflow_template_view,
        'POST': permission_workflow_template_edit
    }
    serializer_class = WorkflowTemplateStateSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'workflow': self.get_workflow_template()
        }

    def get_source_queryset(self):
        return self.get_workflow_template().states.all()


class APIWorkflowTemplateStateView(
    ParentObjectWorkflowTemplateAPIViewMixin,
    generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected workflow template state.
    get: Return the details of the selected workflow template state.
    patch: Edit the selected workflow template state.
    put: Edit the selected workflow template state.
    """
    mayan_object_permission_map = {
        'DELETE': permission_workflow_template_edit,
        'GET': permission_workflow_template_view,
        'PATCH': permission_workflow_template_edit,
        'PUT': permission_workflow_template_edit
    }
    lookup_url_kwarg = 'workflow_template_state_id'
    serializer_class = WorkflowTemplateStateSerializer

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_source_queryset(self):
        return self.get_workflow_template().states.all()


class APIWorkflowTemplateStateActionListView(
    ParentObjectWorkflowTemplateStateAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of all the workflow template state actions.
    post: Create a new workflow template state action.
    """
    mayan_external_object_permission_map = {
        'GET': permission_workflow_template_view,
        'POST': permission_workflow_template_edit
    }
    serializer_class = WorkflowTemplateStateActionSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'state': self.get_workflow_template_state()
        }

    def get_source_queryset(self):
        return self.get_workflow_template_state().actions.all()


class APIWorkflowTemplateStateActionDetailView(
    ParentObjectWorkflowTemplateStateAPIViewMixin,
    generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected workflow template state action.
    get: Return the details of the selected workflow template state action.
    patch: Edit the selected workflow template state action.
    put: Edit the selected workflow template state action.
    """
    mayan_object_permission_map = {
        'DELETE': permission_workflow_template_edit,
        'GET': permission_workflow_template_view,
        'PATCH': permission_workflow_template_edit,
        'PUT': permission_workflow_template_edit
    }
    lookup_url_kwarg = 'workflow_template_state_action_id'
    serializer_class = WorkflowTemplateStateActionSerializer

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}

    def get_source_queryset(self):
        return self.get_workflow_template_state().actions.all()



import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models.signals import post_save
from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig
from mayan.apps.common.menus import menu_list_facet, menu_object
from mayan.apps.events.classes import ModelEventType

from .events import event_user_locale_profile_edited
from .handlers import handler_user_locale_profile_create
from .links import (
    link_user_locale_profile_detail, link_user_locale_profile_edit
)
from .patches import patchDjangoTranslation

logger = logging.getLogger(name=__name__)


class LocalesApp(MayanAppConfig):
    app_namespace = 'locales'
    app_url = 'locales'
    has_rest_api = False
    has_tests = True
    name = 'mayan.apps.locales'
    verbose_name = _(message='Locales')

    def ready(self):
        super().ready()
        User = get_user_model()

        patchDjangoTranslation()

        ModelEventType.register(
            model=User, event_types=(
                event_user_locale_profile_edited,
            )
        )

        menu_list_facet.bind_links(
            links=(
                link_user_locale_profile_detail,
            ), sources=(User,)
        )

        menu_object.bind_links(
            links=(
                link_user_locale_profile_edit,
            ), sources=(User,)
        )

        post_save.connect(
            dispatch_uid='common_handler_user_locale_profile_create',
            receiver=handler_user_locale_profile_create,
            sender=settings.AUTH_USER_MODEL
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.links import Link

from .icons import icon_otp_detail, icon_otp_disable, icon_otp_enable
from .utils import is_otp_backend_enabled


def condition_is_otp_backend_enabled(context, resolved_object):
    return is_otp_backend_enabled()


def condition_is_current_user(context, resolved_object):
    return condition_is_otp_backend_enabled(
        context=context, resolved_object=resolved_object
    ) and context['request'].user == resolved_object and not condition_otp_is_enabled(
        context=context, resolved_object=resolved_object
    )


def condition_otp_is_enabled(context, resolved_object):
    return condition_is_otp_backend_enabled(
        context=context, resolved_object=resolved_object
    ) and resolved_object.otp_data.is_enabled()


link_otp_detail = Link(
    condition=condition_is_otp_backend_enabled, icon=icon_otp_detail,
    text=_(message='OTP details'), view='authentication_otp:otp_detail'
)
link_otp_disable = Link(
    condition=condition_otp_is_enabled, icon=icon_otp_disable,
    text=_(message='Disable OTP'), view='authentication_otp:otp_disable'
)
link_otp_enable = Link(
    condition=condition_is_current_user, icon=icon_otp_enable,
    text=_(message='Enable OTP'), view='authentication_otp:otp_enable'
)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('sources', '0030_documentfilesourcemetadata')
    ]

    operations = [
        migrations.AlterField(
            field=models.TextField(
                blank=True, help_text='The actual value stored in the source '
                'metadata for the document file.', null=True,
                verbose_name='Value'
            ), model_name='documentfilesourcemetadata', name='value'
        )
    ]



from django.utils.translation import gettext_lazy as _

from actstream.models import Action

from mayan.apps.common.serializers import ContentTypeSerializer
from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.fields import DynamicSerializerField

from .event_type_serializers import EventTypeSerializer


class EventSerializer(serializers.ModelSerializer):
    actor = DynamicSerializerField(
        label=_(message='Actor'), read_only=True
    )
    actor_content_type = ContentTypeSerializer(
        label=_(message='Actor content type'), read_only=True
    )
    target = DynamicSerializerField(
        label=_(message='Target'), read_only=True
    )
    target_content_type = ContentTypeSerializer(
        label=_(message='Target content type'), read_only=True
    )
    verb = EventTypeSerializer(
        label=_(message='Verb'), read_only=True
    )

    class Meta:
        exclude = (
            'action_object_content_type', 'action_object_object_id'
        )
        model = Action
        read_only_fields = (
            'action', 'actor_content_type', 'target', 'target_content_type',
            'verb'
        )



from mayan.apps.authentication.settings import setting_authentication_backend
from mayan.apps.common.utils import get_class_full_name


def is_otp_backend_enabled():
    # Hidden import.
    from .authentication_backends import (
        AuthenticationBackendModelUsernamePasswordTOTP
    )

    return setting_authentication_backend.value == get_class_full_name(
        klass=AuthenticationBackendModelUsernamePasswordTOTP
    )



from django.utils.translation import gettext_lazy as _

from mayan.apps.app_manager.apps import MayanAppConfig


class DocumentationApp(MayanAppConfig):
    app_namespace = 'documentation'
    app_url = 'documentation'
    name = 'mayan.apps.documentation'
    verbose_name = _(message='Documentation')



from .callback_mixins import SourceBackendActionMixinCallbackBase


class SourceBackendActionMixinSourceMetadata(
    SourceBackendActionMixinCallbackBase
):
    def get_callback_kwargs_post_document_file_create(self, task_kwargs):
        result = super().get_callback_kwargs_post_document_file_create(
            task_kwargs=task_kwargs
        )

        server_upload_entry = task_kwargs['server_upload_entry']

        result['source_metadata'] = server_upload_entry.get(
            'source_metadata', {}
        )

        return result

    def process_server_upload_entry(self, server_upload_entry):
        result = super().process_server_upload_entry(
            server_upload_entry=server_upload_entry
        )

        source_metadata = server_upload_entry.get(
            'source_metadata', {}
        )

        result['source_metadata'] = source_metadata

        return result



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.menus import Menu
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import icon_cabinet_list
from .permissions import permission_cabinet_create, permission_cabinet_view

menu_cabinets = Menu(
    condition=factory_condition_queryset_access(
        app_label='cabinets', model_name='Cabinet',
        object_permission=permission_cabinet_view,
        view_permission=permission_cabinet_create,
    ), icon=icon_cabinet_list, label=_(message='Cabinets'), name='cabinets'
)



from django.urls import re_path

from .views import (
    CheckVersionView, DependencyGroupEntryListView,
    DependencyGroupEntryDetailView, DependencyGroupListView,
    DependencyLicensesView
)

urlpatterns_groups = [
    re_path(
        route=r'^groups/(?P<dependency_group_name>\w+)/(?P<dependency_group_entry_name>\w+)$',
        name='dependency_group_entry_detail',
        view=DependencyGroupEntryDetailView.as_view()
    ),
    re_path(
        route=r'^groups/(?P<dependency_group_name>\w+)/$',
        name='dependency_group_entry_list',
        view=DependencyGroupEntryListView.as_view()
    ),
    re_path(
        route=r'^groups/$', name='dependency_group_list',
        view=DependencyGroupListView.as_view()
    )
]

urlpatterns_licences = [
    re_path(
        route=r'^licenses/$', name='dependency_licenses_view',
        view=DependencyLicensesView.as_view()
    )
]

urlpatterns_versions = [
    re_path(
        route=r'^version/check/$', name='check_version_view',
        view=CheckVersionView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_groups)
urlpatterns.extend(urlpatterns_licences)
urlpatterns.extend(urlpatterns_versions)



import logging
import re

from django.apps import apps
from django.utils.translation import gettext_lazy as _

from mayan.apps.templating.template_backends import Template

from .literals import (
    REGULAR_EXPRESSION_MATCH_EVERYTHING, REGULAR_EXPRESSION_MATCH_NOTHING
)

logger = logging.getLogger(name=__name__)


class SourceBackendMixinRegularExpression:
    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'include_regex': {
                    'class': 'mayan.apps.templating.fields.TemplateField',
                    'default': '',
                    'kwargs': {
                        'initial_help_text': _(
                            message='Regular expression used to select which '
                            'files to upload.'
                        )
                    },
                    'label': _(message='Include regular expression'),
                    'required': False
                },
                'exclude_regex': {
                    'class': 'mayan.apps.templating.fields.TemplateField',
                    'default': '',
                    'kwargs': {
                        'initial_help_text': _(
                            message='Regular expression used to exclude '
                            'which files to upload.'
                        )
                    },
                    'label': _(message='Exclude regular expression'),
                    'required': False
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='Content selection'), {
                    'fields': ('include_regex', 'exclude_regex')
                }
            ),
        )

        return fieldsets

    def get_regex_exclude(self):
        template_string = self.kwargs.get(
            'exclude_regex', REGULAR_EXPRESSION_MATCH_NOTHING
        ) or REGULAR_EXPRESSION_MATCH_NOTHING

        template = Template(template_string=template_string)

        template_result = template.render()

        return re.compile(pattern=template_result)

    def get_regex_include(self):
        template_string = self.kwargs.get(
            'include_regex', REGULAR_EXPRESSION_MATCH_EVERYTHING
        )

        template = Template(template_string=template_string)

        template_result = template.render()

        return re.compile(pattern=template_result)


class SourceBackendMixinSourceMetadata:
    def callback_post_document_file_create(
        self, source_metadata=None, **kwargs
    ):
        super().callback_post_document_file_create(**kwargs)

        DocumentFileSourceMetadata = apps.get_model(
            app_label='sources', model_name='DocumentFileSourceMetadata'
        )

        document_file = kwargs['document_file']
        source_id = kwargs['source_id']
        source_metadata = source_metadata or {}

        coroutine = DocumentFileSourceMetadata.objects.create_bulk()
        next(coroutine)

        coroutine.send(
            {
                'document_file': document_file, 'key': 'source_id',
                'value': source_id
            }
        )

        for key, value in source_metadata.items():
            coroutine.send(
                {'document_file': document_file, 'key': key, 'value': value}
            )

        coroutine.close()



from mayan.apps.converter.transformations import TransformationResize
from mayan.apps.navigation.column_widgets import SourceColumnWidget

from .settings import setting_thumbnail_height, setting_thumbnail_width


class SourceColumnWidgetDocumentLink(SourceColumnWidget):
    template_name = 'documents/column_widgets/document_link.html'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.document = self.get_document()

    def get_document(self):
        return self.value.document

    def get_extra_context(self):
        return {'instance': self.document}


class ThumbnailWidget(SourceColumnWidget):
    template_name = 'documents/widgets/thumbnail.html'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.document = self.get_document()

    def disable_condition(self):
        return self.document.is_in_trash

    def get_document(self):
        return self.value

    def get_extra_context(self):
        transformation_resize = TransformationResize(
            height=setting_thumbnail_height.value,
            width=setting_thumbnail_width.value
        )

        transformation_instance_list = (transformation_resize,)

        return {
            # Disable the clickable link if the document is in the trash.
            'disable_title_link': self.disable_condition(),
            'gallery_name': 'document_list',
            'instance': self.document,
            'transformation_instance_list': transformation_instance_list
        }



from django.utils.translation import gettext_lazy as _

from mayan.apps.forms import form_fields, form_widgets, forms, formsets


class DocumentDownloadForm(forms.Form):
    document_file_id = form_fields.CharField(
        label=_(message='Document file ID'), widget=form_widgets.HiddenInput
    )
    document = form_fields.CharField(
        label=_(message='Document'), required=False,
        widget=form_widgets.TextInput(
            attrs={
                'readonly': 'readonly'
            }
        )
    )
    document_file = form_fields.CharField(
        label=_(message='Document file'), required=False,
        widget=form_widgets.TextInput(
            attrs={
                'readonly': 'readonly'
            }
        )
    )
    include = form_fields.BooleanField(
        initial=True, label=_(message='Include'), required=False
    )


DocumentDownloadFormSet = formsets.formset_factory(
    form=DocumentDownloadForm, extra=0
)
